#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import warnings
from typing import Any, cast

from pandas import DataFrame

import synnax.channel.payload as channel
from synnax.channel.retrieve import Retriever as ChannelRetriever
from synnax.channel.retrieve import retrieve_required as retrieve_required_channel
from synnax.exceptions import PathError, ValidationError
from synnax.framer.codec import Codec
from synnax.framer.frame import CrudeFrame, Frame
from synnax.telem import DataType
from synnax.telem.series import CrudeSeries, Series


class ReadFrameAdapter:
    __adapter: dict[channel.Key, str] | None
    retriever: ChannelRetriever
    keys: list[channel.Key]
    codec: Codec

    def __init__(self, retriever: ChannelRetriever):
        self.retriever = retriever
        self.__adapter = None
        self.keys = list()
        self.codec = Codec()

    def update(self, channels: channel.Params) -> None:
        normal = channel.normalize_params(channels)
        fetched = self.retriever.retrieve(normal.channels)
        self.codec.update(
            [ch.key for ch in fetched],
            [ch.data_type for ch in fetched],
        )

        if isinstance(normal, channel.NormalizedKeyResult):
            self.__adapter = None
            self.keys = list(normal.channels)
            return

        self.__adapter = dict[int, str]()
        for name in normal.channels:
            ch = next((c for c in fetched if c.name == name), None)
            if ch is None:
                raise KeyError(f"Channel {name} not found.")
            self.__adapter[ch.key] = ch.name
        self.keys = list(self.__adapter.keys())

    def adapt(self, fr: Frame) -> Frame:
        if self.__adapter is None:
            return fr

        # In certain cases there can be a slight desync between the channels being sent
        # by the server and the channels being received by the client. To handle this,
        # we purge any channels that are not in the adapter.
        to_purge: list[int] | None = None
        for i, k in enumerate(fr.channels):
            try:
                if isinstance(k, channel.Key):
                    fr.channels[i] = self.__adapter[k]  # type: ignore[call-overload]
            except KeyError:
                if to_purge is None:
                    to_purge = [i]
                else:
                    to_purge.append(i)
        if to_purge is not None:
            fr.channels = cast(
                list[channel.Key] | list[str],
                [k for i, k in enumerate(fr.channels) if i not in to_purge],
            )
            fr.series = [s for i, s in enumerate(fr.series) if i not in to_purge]

        return fr


class WriteFrameAdapter:
    _adapter: dict[str, channel.Key] | None
    _keys: list[channel.Key]
    _err_on_extra_chans: bool
    _strict_data_types: bool
    _suppress_warnings: bool

    retriever: ChannelRetriever
    codec: Codec

    def __init__(
        self,
        retriever: ChannelRetriever,
        err_on_extra_chans: bool = True,
        strict_data_types: bool = False,
        suppress_warnings: bool = True,
    ):
        self.retriever = retriever
        self._adapter = None
        self._keys = list()
        self._err_on_extra_chans = err_on_extra_chans
        self._strict_data_types = strict_data_types
        self._suppress_warnings = suppress_warnings
        self.codec = Codec()

    def update(self, channels: channel.Params) -> None:
        results = retrieve_required_channel(self.retriever, channels)
        self._adapter = {ch.name: ch.key for ch in results}
        self._keys = [ch.key for ch in results]
        self.codec.update(
            self._keys,
            [ch.data_type for ch in results],
        )

    def adapt_dict_keys(
        self, data: dict[channel.Payload | channel.Key | str, Any]
    ) -> dict[channel.Key, Any]:
        out = dict()
        for k in data.keys():
            out[self.__adapt_to_key(k)] = data[k]
        return out

    @property
    def keys(self) -> list[channel.Key]:
        return self._keys

    def __adapt_to_key(self, ch: channel.Payload | channel.Key | str) -> channel.Key:
        if isinstance(ch, channel.Key):
            return ch
        if isinstance(ch, channel.Payload):
            return ch.key
        # If it's not a payload or key already, it has to be a name,
        # which means we need to resolve the key from a remote source
        # (either cache or server)
        return self.__adapt_ch(ch).key

    def __adapt_ch(self, ch: channel.Key | str | channel.Payload) -> channel.Payload:
        if isinstance(ch, (channel.Key, str)):
            return self.retriever.retrieve_one(ch)
        return ch

    def adapt(
        self,
        channels_or_data: (
            channel.Payload | list[channel.Payload] | channel.Params | CrudeFrame
        ),
        series: CrudeSeries | list[CrudeSeries] | None = None,
    ) -> Frame:
        frame = self._adapt(channels_or_data, series)
        extra = set(frame.channels) - set(self.keys)
        if extra:
            raise PathError("keys", ValidationError(f"frame has extra keys {extra}"))

        for i, (col, series) in enumerate(frame.items()):
            ch = self.retriever.retrieve(col)[0]
            if series.data_type != ch.data_type:
                if self._strict_data_types:
                    raise PathError(
                        str(col),
                        ValidationError(
                            f"Data type {ch.data_type} for channel {ch} does "
                            + f"not match series data type {series.data_type}.",
                        ),
                    )
                elif not self._suppress_warnings and not (
                    ch.data_type == DataType.TIMESTAMP
                    and series.data_type == DataType.INT64
                ):
                    warnings.warn(
                        f"""Series for channel {ch.name} has type {series.data_type} but
                        channel expects type {ch.data_type}. We can safely convert
                        between the two, but this can cause performance degradations
                        and is not recommended. To suppress this warning,
                        set suppress_warnings=True when constructing the writer. To
                        raise an error instead, set strict=True when constructing
                        the writer."""
                    )
                frame.series[i] = series.astype(ch.data_type)
        return frame

    def _adapt(
        self,
        channels_or_data: (
            channel.Payload | list[channel.Payload] | channel.Params | CrudeFrame
        ),
        series: CrudeSeries | list[CrudeSeries] | None = None,
    ) -> Frame:
        if isinstance(channels_or_data, (str, channel.Key, channel.Payload)):
            if series is None:
                raise ValidationError(f"""
                Received a single channel {'name' if isinstance(channels_or_data, str) else 'key'}
                but no data.
                """)
            if isinstance(series, list) and len(series) > 1:
                first = series[0]
                if not isinstance(first, (float, int)):
                    raise ValidationError(f"""
                    Received a single channel {'name' if isinstance(channels_or_data, str) else 'key'}
                    but multiple series.
                    """)

            pld = self.__adapt_ch(channels_or_data)
            return Frame([pld.key], [Series(cast(CrudeSeries, series))])

        if isinstance(channels_or_data, list):
            if series is None:
                raise ValidationError(f"""
                Received {len(channels_or_data)} channels but no series.
                """)
            series_list: list[CrudeSeries] = (
                [series]
                if not isinstance(series, list)
                else cast(list[CrudeSeries], series)
            )
            channels: list[channel.Key] = list()
            o_series: list[Series] = list()
            for i, ch in enumerate(channels_or_data):
                pld = self.__adapt_ch(ch)
                if i >= len(series_list):
                    raise ValidationError(f"""
                    Received {len(channels_or_data)} channels but only {len(series_list)} series.
                    """)
                channels.append(pld.key)
                o_series.append(Series(series_list[i]))

            return Frame(channels, o_series)

        if isinstance(channels_or_data, (Frame, DataFrame)):
            if isinstance(channels_or_data, Frame):
                cols: list[channel.Key] | list[str] = channels_or_data.channels
            else:
                cols = cast(list[str], list(channels_or_data.columns))
            if self._adapter is None:
                return (
                    Frame(channels_or_data)
                    if isinstance(channels_or_data, DataFrame)
                    else channels_or_data
                )
            adapted_channels: list[channel.Key] = list()
            adapted_series: list[Series] = list()
            for col in cols:
                try:
                    adapted_channels.append(
                        self._adapter[col] if isinstance(col, str) else col
                    )
                    adapted_series.append(Series(channels_or_data[col]))
                except KeyError as e:
                    if self._err_on_extra_chans:
                        raise ValidationError(
                            f"Channel {e} was not provided in the list of "
                            f"channels when the writer was opened."
                        )
            return Frame(channels=adapted_channels, series=adapted_series)

        if isinstance(channels_or_data, dict):
            dict_channels: list[channel.Key] = list()
            dict_series: list[Series] = list()
            for k, v in channels_or_data.items():
                pld = self.__adapt_ch(k)
                dict_channels.append(pld.key)
                dict_series.append(Series(v))

            return Frame(dict_channels, dict_series)

        raise TypeError(
            f"""Cannot construct frame from {channels_or_data} and {series}"""
        )



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import synnax as sy

"""
This example demonstrates how to configure and start a Counter Read Task on a National
Instruments USB-6289 device to measure frequency and count edges.

To run this example, you'll need to have your Synnax cluster properly configured to
detect National Instruments devices: https://docs.synnaxlabs.com/reference/driver/ni/get-started

You'll also need to have either a Counter Input module or create a simulated device
via the NI-MAX software.
"""

# We've logged in via the CLI, so there's no need to provide credentials here.
# See https://docs.synnaxlabs.com/reference/client/quick-start for more information.
client = sy.Synnax()

dev = client.devices.retrieve(model="NI 9326")

# Create an index channel that will be used to store the timestamps
# for the counter read data.
ci_time = client.channels.create(
    name="ci_time",
    is_index=True,
    data_type=sy.DataType.TIMESTAMP,
    retrieve_if_name_exists=True,
)

# Create synnax channels that will be used to store the counter input data.
# We'll create one for frequency measurement and one for edge counting.
ci_0_freq = client.channels.create(
    name="ci_0_freq",
    # Pass in the index key here to associate the channel with the index channel.
    index=ci_time.key,
    data_type=sy.DataType.FLOAT64,
    retrieve_if_name_exists=True,
)

ci_1_count = client.channels.create(
    name="ci_1_count",
    # Pass in the index key here to associate the channel with the index channel.
    index=ci_time.key,
    data_type=sy.DataType.UINT32,
    retrieve_if_name_exists=True,
)

# Instantiate the task. A task is a background process that can be used to acquire data
# from, or write commands to a device. Tasks are the primary method for interacting with
# hardware in Synnax.
tsk = sy.ni.CounterReadTask(
    # A name to find and monitor the task via the Synnax Console.
    name="Basic Counter Read",
    # The key of the device to execute the task on.
    device=dev.key,
    # The rate at which the task will sample data from the device.
    sample_rate=sy.Rate.HZ * 100,
    # The rate at which data will be streamed from the device into Synnax.
    # Since we're sampling at 100 Hz and streaming at 25 Hz, we'll get 4 samples at a
    # time.
    # It's generally best to keep the stream rate under 100 Hz.
    stream_rate=sy.Rate.HZ * 25,
    # Whether to save data acquired by the task to disk. If set to False, the data
    # will be streamed into Synnax for real-time consumption but not saved to disk.
    data_saving=True,
    # The list of counter channels we'd like to acquire data from.
    channels=[
        sy.ni.CIFrequencyChan(
            # The key of the Synnax channel we're acquiring data for.
            channel=ci_0_freq.key,
            # The counter port on the device (ctr0, ctr1, etc.)
            port=0,
            # Minimum expected frequency value in Hz
            min_val=1,
            # Maximum expected frequency value in Hz
            max_val=10000,
            # Units of measurement (Hz, Seconds, or Ticks)
            units="Hz",
            # Edge to count on (Rising or Falling)
            edge="Rising",
            # Measurement method (LowFreq1Ctr, HighFreq2Ctr, or DynAvg)
            meas_method="LowFreq1Ctr",
            # Measurement time in seconds
            meas_time=0.001,
            # Divisor for internal timebase
            divisor=4,
        ),
        sy.ni.CIEdgeCountChan(
            # The key of the Synnax channel we're acquiring data for.
            channel=ci_1_count.key,
            # The counter port on the device
            port=1,
            # Edge to count on (Rising or Falling)
            active_edge="Rising",
            # Count direction (CountUp, CountDown, or ExtControlled)
            count_direction="CountUp",
            # Initial count value
            initial_count=0,
        ),
    ],
)

# This will create the task in Synnax and wait for the driver to validate that the
# configuration is correct.
client.tasks.configure(tsk)

# Stream 100 reads, which will accumulate a total of 400 samples
# for each channel over a period of 4 seconds.
total_reads = 100

# Create a synnax frame to accumulate data.
frame = sy.Frame()

# Start the task under a context manager, which ensures the task gets stopped
# when the block exits.
with tsk.run():
    # Open a streamer on the counter input channels.
    with client.open_streamer(["ci_0_freq", "ci_1_count"]) as streamer:
        for i in range(total_reads):
            frame.append(streamer.read())

# Clean up by deleting the task
client.tasks.delete(tsk.key)

# Save the data to a CSV file.
frame.to_df().to_csv("counter_read_result.csv")



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import warnings

from alamos import NOOP, Instrumentation
from freighter import URL

from synnax import (
    access,
    arc,
    auth,
    channel,
    control,
    device,
    framer,
    group,
    ontology,
    rack,
    ranger,
    signals,
    status,
    task,
    user,
)
from synnax.config import try_load_options_if_none_provided
from synnax.options import Options
from synnax.telem import TimeSpan
from synnax.transport import Transport


class Synnax(framer.Client):
    """Client to perform operations against a Synnax cluster.

    If using the python client for data analysis/personal use, the easiest way to
    connect is to use the `synnax login` command, which will prompt and securely
    store your credentials. The client can then be initialized without parameters. When
    using the client in a production environment, it's best to provide the connection
    parameter as arguments loaded from a configuration or environment variable.

    After running the synnax login command::
        client = Synnax()

    Without running the synnax login command::
        client = Synnax(
            host="synnax.example.com",
            port=9090,
            username="synnax",
            password="seldon",
            secure=True
        )
    """

    channels: channel.Client
    access: access.Client
    users: user.Client
    ranges: ranger.Client
    control: control.Client
    signals: signals.Registry
    racks: rack.Client
    devices: device.Client
    tasks: task.Client
    ontology: ontology.Client
    statuses: status.Client
    arcs: arc.Client
    groups: group.Client

    _transport: Transport

    def __init__(
        self,
        host: str = "",
        port: int = 0,
        username: str = "",
        password: str = "",
        secure: bool = False,
        open_timeout: TimeSpan = TimeSpan.SECOND * 5,
        read_timeout: TimeSpan = TimeSpan.SECOND * 5,
        keep_alive: TimeSpan = TimeSpan.SECOND * 30,
        max_retries: int = 3,
        instrumentation: Instrumentation = NOOP,
        cache_channels: bool = True,
    ):
        """Creates a new client. Connection parameters can be provided as arguments, or,
        if none are provided, the client will attempt to load them from the Synnax
        configuration file (~/.synnax/config.json) as well as credentials stored in the
        operating system's keychain.

        If using the client for data analysis/personal use, the easiest way to connect
        is to use the `synnax login` command, which will prompt and securely store your
        credentials. The client can then be initialized without parameters.

        :param host: Hostname of a node in the Synnax cluster.
        :param port: Port of the node.
        :param username: Username to authenticate with.
        :param password: Password to authenticate with.
        :param secure: Whether to use TLS when connecting to the cluster.
        """
        opts = try_load_options_if_none_provided(host, port, username, password, secure)
        self._transport = _configure_transport(
            opts=opts,
            open_timeout=open_timeout,
            read_timeout=read_timeout,
            keep_alive=keep_alive,
            max_retries=max_retries,
            instrumentation=instrumentation,
        )
        self.auth = auth.Client(
            transport=self._transport.unary,
            username=opts.username,
            password=opts.password,
        )
        self.auth.authenticate()
        self._transport.use(self.auth.middleware())
        self._transport.use_async(self.auth.async_middleware())

        cluster_retriever = channel.ClusterRetriever(
            self._transport.unary, instrumentation
        )
        cache_retriever: channel.CacheRetriever | None = None
        ch_retriever: channel.Retriever
        if cache_channels:
            cache_retriever = channel.CacheRetriever(cluster_retriever, instrumentation)
            ch_retriever = cache_retriever
        else:
            ch_retriever = cluster_retriever
        deleter = framer.Deleter(self._transport.unary, instrumentation)
        ch_creator = channel.Writer(
            self._transport.unary,
            instrumentation,
            cache_retriever,
        )
        super().__init__(
            stream_client=self._transport.stream,
            async_client=self._transport.stream_async,
            unary_client=self._transport.unary,
            retriever=ch_retriever,
            deleter=deleter,
            instrumentation=instrumentation,
        )
        self.groups = group.Client(self._transport.unary)
        self.ontology = ontology.Client(client=self._transport.unary)
        self.channels = channel.Client(self, ch_retriever, ch_creator)
        range_retriever = ranger.Retriever(self._transport.unary, instrumentation)
        range_creator = ranger.Writer(self._transport.unary, instrumentation)
        self.signals = signals.Registry(frame_client=self, channels=ch_retriever)
        self.racks = rack.Client(client=self._transport.unary)
        self.devices = device.Client(client=self._transport.unary)
        self.tasks = task.Client(
            client=self._transport.unary,
            frame_client=self,
            rack_client=self.racks,
            device_client=self.devices,
        )
        self.ranges = ranger.Client(
            unary_client=self._transport.unary,
            frame_client=self,
            channel_retriever=ch_retriever,
            writer=range_creator,
            retriever=range_retriever,
            signals=self.signals,
            ontology=self.ontology,
            tasks=self.tasks,
        )
        self.control = control.Client(self, ch_retriever)
        self.users = user.Client(self._transport.unary)
        self.statuses = status.Client(self._transport.unary)
        self.arcs = arc.Client(self._transport.unary)
        self.access = access.Client(self._transport.unary)

    @property
    def user(self) -> user.Client:
        """Deprecated: Use users instead."""
        return self.users

    @property
    def hardware(self) -> "Synnax":
        """Deprecated: Use client.devices, client.tasks, client.racks directly."""
        warnings.warn(
            "client.hardware is deprecated and will be removed in a future version. "
            "Use client.devices, client.tasks, client.racks directly instead.",
            FutureWarning,
            stacklevel=2,
        )
        return self

    def close(self) -> None:
        """Shuts down the client and closes all connections. All open iterators or
        writers must be closed before calling this method.
        """
        # No-op for now, we'll definitely add cleanup logic in the future, so it's
        # good to have this API defined.
        ...


def _configure_transport(
    opts: Options,
    open_timeout: TimeSpan,
    read_timeout: TimeSpan,
    keep_alive: TimeSpan,
    max_retries: int,
    instrumentation: Instrumentation = NOOP,
) -> Transport:
    t = Transport(
        instrumentation=instrumentation,
        url=URL(host=opts.host, port=opts.port),
        secure=opts.secure,
        open_timeout=open_timeout,
        read_timeout=read_timeout,
        keep_alive=keep_alive,
        max_retries=max_retries,
    )
    return t



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

import warnings
from collections.abc import Generator
from contextlib import contextmanager
from typing import TYPE_CHECKING, Annotated, Any
from typing import Protocol as BaseProtocol
from typing import overload
from uuid import uuid4

from alamos import NOOP, Instrumentation
from freighter import Empty, UnaryClient, send_required
from pydantic import BaseModel, Field, ValidationError, field_validator

from synnax.device import Client as DeviceClient
from synnax.device import Device
from synnax.exceptions import ConfigurationError, UnexpectedError
from synnax.framer import Client as FrameClient
from synnax.ontology.payload import ID
from synnax.rack import Client as RackClient
from synnax.rack import Rack
from synnax.status import VARIANT_ERROR, VARIANT_SUCCESS
from synnax.task.payload import Payload, Status, ontology_id
from synnax.telem import TimeSpan, TimeStamp
from synnax.util.normalize import check_for_none, normalize, override


class _CreateRequest(BaseModel):
    tasks: list[Payload]


_CreateResponse = _CreateRequest


class _DeleteRequest(BaseModel):
    keys: list[int]


class _CopyRequest(BaseModel):
    key: int
    name: str
    snapshot: bool


class _CopyResponse(BaseModel):
    task: Payload


class _RetrieveRequest(BaseModel):
    rack: int | None = None
    keys: list[int] | None = None
    names: list[str] | None = None
    types: list[str] | None = None
    include_status: bool = False
    internal: bool | None = None
    snapshot: bool | None = None


class _RetrieveResponse(BaseModel):
    tasks: list[Payload] | None = None


_CREATE_ENDPOINT = "/task/create"
_DELETE_ENDPOINT = "/task/delete"
_RETRIEVE_ENDPOINT = "/task/retrieve"
_COPY_ENDPOINT = "/task/copy"

_TASK_STATE_CHANNEL = "sy_status_set"
_TASK_CMD_CHANNEL = "sy_task_cmd"


class BaseConfig(BaseModel):
    """
    Base configuration shared by all hardware task types.

    This base class provides common fields that all hardware integration tasks need:
    auto-start behavior.
    """

    auto_start: bool = False


class BaseReadConfig(BaseConfig):
    """
    Base configuration for hardware read/acquisition tasks.

    Extends BaseTaskConfig with sample rate and stream rate fields common to
    all data acquisition tasks (LabJack, NI, Modbus, OPC UA read tasks).

    Default rate limits are set to 50 kHz based on NI hardware constraints,
    which are the most restrictive across supported hardware platforms.
    Hardware-specific configs can override these limits for devices that
    support higher rates.
    """

    data_saving: bool = True
    "Whether to persist acquired data to disk (True) or only stream it (False)."
    sample_rate: Annotated[int, Field(ge=0, le=50000)]
    "The rate at which to sample data from the hardware device (Hz)."
    stream_rate: Annotated[int, Field(ge=0, le=50000)]
    "The rate at which acquired data will be streamed to the Synnax cluster (Hz)."

    @field_validator("stream_rate")
    def validate_stream_rate(cls, v: int, info: Any) -> int:
        """Validate that stream_rate is less than or equal to sample_rate."""
        if "sample_rate" in info.data and v > info.data["sample_rate"]:
            raise ValueError(
                "Stream rate must be less than or equal to the sample rate"
            )
        return v


class BaseWriteConfig(BaseConfig):
    """
    Base configuration for hardware write/control tasks.

    Provides common fields (device, auto_start) for all hardware write tasks.
    Note that state_rate and data_saving are NOT included in this base class as they
    are hardware-specific - only write tasks with state feedback (NI, LabJack) use
    these fields. Tasks without state feedback (Modbus, OPC UA) do not need them.
    """

    device: str = Field(min_length=1)
    "The key of the Synnax device this task will communicate with."


class Task:
    key: int = 0
    name: str = ""
    type: str = ""
    config: dict[str, Any] = {}
    snapshot: bool = False
    status: Status | None = None
    __frame_client: FrameClient | None = None

    def __init__(
        self,
        *,
        key: int = 0,
        rack: int = 0,
        name: str = "",
        type: str = "",
        config: dict[str, Any] | None = None,
        snapshot: bool = False,
        status: Status | None = None,
        _frame_client: FrameClient | None = None,
    ):
        if key == 0:
            key = (rack << 32) + 0
        self.key = key
        self.name = name
        self.type = type
        self.config = config if config is not None else {}
        self.snapshot = snapshot
        self.status = status
        self.__frame_client = _frame_client

    @property
    def _frame_client(self) -> FrameClient:
        if self.__frame_client is None:
            raise RuntimeError(
                "Cannot execute commands on a task that has not been created or retrieved from the cluster."
            )
        return self.__frame_client

    def to_payload(self) -> Payload:
        return Payload(
            key=self.key,
            name=self.name,
            type=self.type,
            config=self.config,
        )

    def set_internal(self, task: Task) -> None:
        self.key = task.key
        self.name = task.name
        self.type = task.type
        self.config = task.config
        self.snapshot = task.snapshot
        self.__frame_client = task.__frame_client

    @property
    def ontology_id(self) -> ID:
        return ontology_id(self.key)

    def update_device_properties(self, device_client: DeviceClient) -> Device | None:
        """Update device properties before task configuration.

        Default implementation for base Task class does nothing and returns None.
        Tasks that need to update device properties (LabJack, Modbus, OPC UA)
        should override this method in their respective classes.

        Returns:
            None - base implementation performs no updates.
        """
        return None

    def execute_command(self, type_: str, args: dict[str, Any] | None = None) -> str:
        """Executes a command on the task and returns the unique key assigned to the
        command.

        :param type_: The type of command to execute.
        :param args: The arguments to pass to the command.
        :return: The unique key assigned to the command.
        """
        w = self._frame_client.open_writer(TimeStamp.now(), _TASK_CMD_CHANNEL)
        key = str(uuid4())
        w.write(
            _TASK_CMD_CHANNEL,
            [{"task": self.key, "type": type_, "key": key, "args": args}],
        )
        w.close()
        return str(key)

    def execute_command_sync(
        self,
        type_: str,
        args: dict[str, Any] | None = None,
        timeout: float | TimeSpan = 5,
    ) -> Status:
        """Executes a command on the task and waits for the driver to acknowledge the
        command with a state.

        :param type_: The type of command to execute.
        :param args: The arguments to pass to the command.
        :param timeout: The maximum time to wait for the driver to acknowledge the
        command before a timeout occurs.
        """
        with self._frame_client.open_streamer([_TASK_STATE_CHANNEL]) as s:
            key = self.execute_command(type_, args)
            while True:
                frame = s.read(TimeSpan.from_seconds(timeout).seconds)
                if frame is None:
                    raise TimeoutError(
                        f"timed out waiting for driver to acknowledge {type_} command"
                    )
                elif _TASK_STATE_CHANNEL not in frame:
                    warnings.warn("task - unexpected missing state in frame")
                    continue
                try:
                    status = Status.model_validate(frame[_TASK_STATE_CHANNEL][0])
                    if (
                        status.details is not None
                        and status.details.cmd is not None
                        and status.details.cmd == key
                    ):
                        return status
                except ValidationError as e:
                    raise UnexpectedError(f"""
                    Received invalid task state from driver.
                    """) from e


class Protocol(BaseProtocol):
    @property
    def key(self) -> int: ...

    def to_payload(self) -> Payload: ...

    def set_internal(self, task: Task) -> None: ...

    def update_device_properties(self, device_client: DeviceClient) -> Device | None:
        """Update device properties before task configuration.

        This method can be overridden by tasks that need to synchronize
        their configuration with device properties (e.g., Modbus, OPC UA, LabJack).
        The default implementation does nothing.

        Args:
            device_client: Client for accessing device operations

        Returns:
            The updated device, or None if no update was performed.
        """
        ...


class StarterStopperMixin:
    _internal: Task

    def start(self, timeout: float | TimeSpan = 5) -> None:
        """Starts the task and blocks until the Synnax cluster has acknowledged the
        command or the specified timeout has elapsed.

        :raises TimeoutError: If the timeout is reached before the Synnax cluster
            acknowledges the command.
        :raises Exception: If the Synnax cluster fails to start the task correctly.
        """
        self._internal.execute_command_sync("start", timeout=timeout)

    def stop(self, timeout: float | TimeSpan = 5) -> None:
        """Stops the task and blocks until the Synnax cluster has acknowledged the
        command or the specified timeout has elapsed.

        :raises TimeoutError: If the timeout is reached before the Synnax cluster
            acknowledges the command.
        :raises Exception: If the Synnax cluster fails to stop the task correctly.
        """
        self._internal.execute_command_sync("stop", timeout=timeout)

    @contextmanager
    def run(self, timeout: float | TimeSpan = 5) -> Generator[None, None, None]:
        """Context manager that starts the task before entering the block and stops the
        task after exiting the block. This is useful for ensuring that the task is
        properly stopped even if an exception occurs during execution.
        """
        self.start(timeout)
        try:
            yield
        finally:
            self.stop(timeout)


class JSONConfigMixin(Protocol):
    _internal: Task
    config: BaseModel

    @property
    def name(self) -> str:
        return self._internal.name

    @property
    def key(self) -> int:
        """Implements TaskProtocol protocol"""
        return self._internal.key

    def to_payload(self) -> Payload:
        """Implements TaskProtocol protocol"""
        pld = self._internal.to_payload()
        pld.config = self.config.model_dump()
        return pld

    def set_internal(self, task: Task) -> None:
        """Implements TaskProtocol protocol"""
        self._internal = task


class Client:
    _client: UnaryClient
    _frame_client: FrameClient
    _default_rack: Rack | None
    _racks: RackClient
    _device_client: DeviceClient | None
    instrumentation: Instrumentation = NOOP

    def __init__(
        self,
        client: UnaryClient,
        frame_client: FrameClient,
        rack_client: RackClient,
        device_client: DeviceClient | None = None,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self._client = client
        self._frame_client = frame_client
        self._racks = rack_client
        self._device_client = device_client
        self._default_rack = None
        self.instrumentation = instrumentation

    @overload
    def create(
        self,
        *,
        key: int = 0,
        name: str = "",
        type: str = "",
        config: dict[str, Any] | None = None,
        rack: int = 0,
    ) -> Task: ...

    @overload
    def create(self, tasks: Task) -> Task: ...

    @overload
    def create(self, tasks: list[Task]) -> list[Task]: ...

    def create(
        self,
        tasks: Task | list[Task] | None = None,
        *,
        key: int = 0,
        name: str = "",
        type: str = "",
        config: dict[str, Any] | None = None,
        rack: int = 0,
    ) -> Task | list[Task]:
        is_single = True
        if tasks is None:
            payloads = [Payload(key=key, name=name, type=type, config=config or {})]
        elif isinstance(tasks, Task):
            payloads = [tasks.to_payload()]
        else:
            is_single = False
            payloads = [t.to_payload() for t in tasks]
        for pld in payloads:
            self.maybe_assign_def_rack(pld, rack)
        req = _CreateRequest(tasks=payloads)
        created = self.__exec_create(req)
        sugared = self.sugar(created)
        return sugared[0] if is_single else sugared

    def __exec_create(self, req: _CreateRequest) -> list[Payload]:
        res = send_required(self._client, "/task/create", req, _CreateResponse)
        return res.tasks

    def maybe_assign_def_rack(self, pld: Payload, rack: int = 0) -> Payload:
        if self._default_rack is None:
            # Hardcoded as this value for now. Will be changed once we have multi-rack
            # systems
            self._default_rack = self._racks.retrieve_embedded_rack()
        if pld is not None and pld.key == 0:
            if rack == 0:
                rack = self._default_rack.key
            pld.key = (rack << 32) + 0
        return pld

    def configure(self, task: Protocol, timeout: float = 5) -> Protocol:
        # Call task-specific device property update (e.g., for Modbus, OPC UA, LabJack)
        if self._device_client is not None:
            task.update_device_properties(self._device_client)

        with self._frame_client.open_streamer([_TASK_STATE_CHANNEL]) as streamer:
            pld = self.maybe_assign_def_rack(task.to_payload())
            req = _CreateRequest(tasks=[pld])
            tasks = self.__exec_create(req)
            task.set_internal(self.sugar(tasks)[0])
            while True:
                frame = streamer.read(timeout)
                if frame is None:
                    raise TimeoutError(
                        "task - timeout waiting for driver to "
                        "acknowledge configuration"
                    )
                elif (
                    _TASK_STATE_CHANNEL not in frame
                    or len(frame[_TASK_STATE_CHANNEL]) == 0
                ):
                    warnings.warn("task - unexpected missing state in frame")
                    continue
                status = Status.model_validate(frame[_TASK_STATE_CHANNEL][0])
                if status.details is None or status.details.task != task.key:
                    continue
                if status.variant == VARIANT_SUCCESS:
                    break
                if status.variant == VARIANT_ERROR:
                    raise ConfigurationError(status.message)
        return task

    def delete(self, keys: int | list[int]) -> None:
        req = _DeleteRequest(keys=normalize(keys))
        send_required(self._client, "/task/delete", req, Empty)

    @overload
    def retrieve(
        self,
        key: int | None = None,
        name: str | None = None,
        type: str | None = None,
    ) -> Task: ...

    @overload
    def retrieve(
        self,
        key: None = None,
        name: None = None,
        type: None = None,
        names: list[str] | None = None,
        keys: list[int] | None = None,
        types: list[str] | None = None,
    ) -> list[Task]: ...

    def retrieve(
        self,
        key: int | None = None,
        name: str | None = None,
        type: str | None = None,
        names: list[str] | None = None,
        keys: list[int] | None = None,
        types: list[str] | None = None,
    ) -> list[Task] | Task:
        is_single = check_for_none(names, keys, types)
        res = send_required(
            self._client,
            "/task/retrieve",
            _RetrieveRequest(
                keys=override(key, keys),
                names=override(name, names),
                types=override(type, types),
            ),
            _RetrieveResponse,
        )
        sug = self.sugar(res.tasks or [])

        # Warn if multiple tasks found when retrieving by name
        if is_single and name is not None and len(sug) > 1:
            task_keys = ", ".join(str(t.key) for t in sug)
            warnings.warn(
                f"Multiple tasks ({len(sug)}) found with name '{name}'. "
                f"Keys: [{task_keys}]. Returning the first task ({sug[0].key}).",
                UserWarning,
                stacklevel=2,
            )

        return sug[0] if is_single else sug

    def sugar(self, tasks: list[Payload]) -> list[Task]:
        return [Task(**t.model_dump(), _frame_client=self._frame_client) for t in tasks]

    def list(self, rack: int | None = None) -> list[Task]:
        """Lists all tasks on a rack. If no rack is specified, lists all tasks on the
        default rack. Excludes internal system tasks (scanner tasks and rack state).

        :param rack: The rack key to list tasks from. If None, uses the default rack.
        :return: A list of all user-created tasks on the specified rack.
        """
        if rack is None:
            if self._default_rack is None:
                self._default_rack = self._racks.retrieve_embedded_rack()
            rack = self._default_rack.key

        res = send_required(
            self._client,
            _RETRIEVE_ENDPOINT,
            _RetrieveRequest(rack=rack, internal=False),
            _RetrieveResponse,
        )
        return self.sugar(res.tasks or [])

    def copy(
        self,
        key: int,
        name: str,
    ) -> Task:
        """Copies an existing task with a new name.

        :param key: The key of the task to copy.
        :param name: The name for the new task.
        :return: The newly created task.
        """
        req = _CopyRequest(key=key, name=name, snapshot=False)
        res = send_required(self._client, _COPY_ENDPOINT, req, _CopyResponse)
        return self.sugar([res.task])[0]



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import functools
from collections.abc import Callable
from typing import ParamSpec, TypeVar, overload

P = ParamSpec("P")
R = TypeVar("R")


class RequiresNamedParams(TypeError):
    """
    Custom exception raised when a function is called with positional arguments
    instead of named arguments.
    """

    pass


@overload
def require_named_params(func: Callable[P, R]) -> Callable[P, R]: ...


@overload
def require_named_params(
    func: None = None, *, example_params: tuple[str, str] | None = None
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...


def require_named_params(
    func: Callable[P, R] | None = None,
    *,
    example_params: tuple[str, str] | None = None,
) -> Callable[P, R] | Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Decorator that catches TypeError exceptions related to positional arguments
    and re-raises them with a more helpful error message.

    Args:
        func: The function to decorate
        example_params: Optional tuple of (param_name, param_value) to show in the
            error message. Example: example_params=("user_id", "12345")

    Returns:
        The decorated function with improved error messages for positional arguments
    """

    def decorator(f: Callable[P, R]) -> Callable[P, R]:
        @functools.wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            try:
                return f(*args, **kwargs)
            except TypeError as e:
                if "positional argument" in str(e) and "were given" in str(e):
                    func_name = f.__qualname__
                    if example_params:
                        param_name, param_value = example_params
                        param_example = f"{func_name}({param_name}='{param_value}')"
                        value_example = f"'{param_value}'"
                    else:
                        param_example = f"{func_name}(name='value')"
                        value_example = "'value'"
                    message = (
                        f"{str(e)}. '{func_name}' only accepts named"
                        f" parameters.\nTry using named parameters"
                        f" like: {param_example} instead of"
                        f" {func_name}({value_example})"
                    )
                    raise RequiresNamedParams(message) from None
                raise

        return wrapper

    if func is None:
        return decorator
    return decorator(func)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import serial

import synnax as sy

PORT = "/dev/cu.usbmodem21401"
BAUD_RATE = 9600

# Create the Synnax client
client = sy.Synnax(
    host="localhost",
    port=9090,
    username="synnax",
    password="seldon",
)

# Create the index channel
index_channel = client.channels.create(
    name="arduino_time",
    is_index=True,
    data_type="timestamp",
    retrieve_if_name_exists=True,
)

# Create the data channel
data_channel = client.channels.create(
    name="arduino_value",
    index=index_channel.key,
    data_type="float32",
    retrieve_if_name_exists=True,
)

# Set up the serial connection
ser = serial.Serial(PORT, BAUD_RATE)
if ser.is_open:
    print("Serial connection established")
else:
    print("Failed to establish serial connection")

# Open a writer and continuously read from the Arduino
with client.open_writer(
    start=sy.TimeStamp.now(),
    channels=["arduino_time", "arduino_value"],
) as writer:
    while True:
        # Read from the serial connection
        value = float(ser.readline().decode("utf-8").rstrip())
        print(value)
        writer.write({"arduino_time": sy.TimeStamp.now(), "arduino_value": value})



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.


import numpy as np
import pytest

import synnax as sy
from synnax.channel.payload import Key, Params
from synnax.framer.codec import Codec


@pytest.mark.framer
@pytest.mark.frame_codec
class TestCodec:
    class Spec:
        def __init__(
            self,
            name: str,
            channels: list[Key] | tuple[Key],
            data_types: list[sy.DataType],
            frame: sy.Frame,
        ):
            self.name = name
            self.channels = channels
            self.data_types = data_types
            self.frame = frame

    @pytest.mark.parametrize(
        "spec",
        [
            Spec(
                name="All Channels Present, In Order",
                channels=[1, 2, 3],
                data_types=[
                    sy.DataType.INT64,
                    sy.DataType.FLOAT32,
                    sy.DataType.FLOAT64,
                ],
                frame=sy.Frame(
                    channels=[1, 2, 3],
                    series=[
                        sy.Series(data=np.array([1, 2, 3], dtype=np.int64)),
                        sy.Series(data=np.array([4, 5, 6], dtype=np.float32)),
                        sy.Series(data=np.array([7, 8, 9], dtype=np.float64)),
                    ],
                ),
            ),
            Spec(
                name="All Channels Present, Out of Order",
                channels=[3, 1, 2],
                data_types=[
                    sy.DataType.FLOAT64,
                    sy.DataType.INT64,
                    sy.DataType.FLOAT32,
                ],
                frame=sy.Frame(
                    channels=[2, 3, 1],
                    series=[
                        sy.Series(data=np.array([4, 5, 6], dtype=np.float32)),
                        sy.Series(data=np.array([7, 8, 9], dtype=np.float64)),
                        sy.Series(data=np.array([1, 2, 3], dtype=np.int64)),
                    ],
                ),
            ),
            Spec(
                name="Some Channels Present, In Order",
                channels=[1, 2, 3],
                data_types=[
                    sy.DataType.UINT8,
                    sy.DataType.FLOAT32,
                    sy.DataType.FLOAT64,
                ],
                frame=sy.Frame(
                    channels=[1, 3],
                    series=[
                        sy.Series(data=np.array([1, 2, 3], dtype=np.uint8)),
                        sy.Series(data=np.array([7, 8, 9], dtype=np.float64)),
                    ],
                ),
            ),
            Spec(
                name="Some Channels Present, Out of Order",
                channels=[1, 2, 3],
                data_types=[
                    sy.DataType.UINT8,
                    sy.DataType.FLOAT32,
                    sy.DataType.FLOAT64,
                ],
                frame=sy.Frame(
                    channels=[3, 1],
                    series=[
                        sy.Series(data=np.array([7, 8, 9], dtype=np.float64)),
                        sy.Series(data=np.array([1, 2, 3], dtype=np.uint8)),
                    ],
                ),
            ),
            Spec(
                name="Only One Channel Present",
                channels=[1, 2, 3, 4, 5],
                data_types=[
                    sy.DataType.UINT8,
                    sy.DataType.UINT8,
                    sy.DataType.UINT8,
                    sy.DataType.UINT8,
                    sy.DataType.UINT8,
                ],
                frame=sy.Frame(
                    channels=[3],
                    series=[
                        sy.Series(data=np.array([1, 2, 3, 4, 5], dtype=np.uint8)),
                    ],
                ),
            ),
            Spec(
                name="All Same Time Range",
                channels=[1, 2],
                data_types=[sy.DataType.UINT8, sy.DataType.FLOAT32],
                frame=sy.Frame(
                    channels=[1, 2],
                    series=[
                        sy.Series(
                            data_type=sy.DataType.UINT8,
                            data=np.array([1], dtype=np.uint8),
                            time_range=sy.TimeRange(start=0, end=5),
                        ),
                        sy.Series(
                            data_type=sy.DataType.FLOAT32,
                            data=np.array([1, 2, 3, 4], dtype=np.float32),
                            time_range=sy.TimeRange(start=0, end=5),
                        ),
                    ],
                ),
            ),
            Spec(
                name="Different Time Ranges",
                channels=[1, 2],
                data_types=[sy.DataType.UINT8, sy.DataType.FLOAT32],
                frame=sy.Frame(
                    channels=[1, 2],
                    series=[
                        sy.Series(
                            data_type=sy.DataType.UINT8,
                            data=np.array([1], dtype=np.uint8),
                            time_range=sy.TimeRange(start=0, end=5),
                        ),
                        sy.Series(
                            data_type=sy.DataType.FLOAT32,
                            data=np.array([1, 2, 3, 4], dtype=np.float32),
                            time_range=sy.TimeRange(start=5, end=10),
                        ),
                    ],
                ),
            ),
            Spec(
                name="Partial Present, Different Lengths",
                channels=[1, 2, 3],
                data_types=[
                    sy.DataType.UINT8,
                    sy.DataType.FLOAT32,
                    sy.DataType.FLOAT64,
                ],
                frame=sy.Frame(
                    channels=[1, 3],
                    series=[
                        sy.Series(data=np.array([1], dtype=np.uint8)),
                        sy.Series(data=np.array([1, 2, 3, 4], dtype=np.float64)),
                    ],
                ),
            ),
            Spec(
                name="Same Alignments",
                channels=[1, 2],
                data_types=[sy.DataType.UINT8, sy.DataType.FLOAT32],
                frame=sy.Frame(
                    channels=[1, 2],
                    series=[
                        sy.Series(
                            data_type=sy.DataType.UINT8,
                            data=np.array([1], dtype=np.uint8),
                            alignment=5,
                        ),
                        sy.Series(
                            data_type=sy.DataType.FLOAT32,
                            data=np.array([1, 2, 3, 4], dtype=np.float32),
                            alignment=5,
                        ),
                    ],
                ),
            ),
            Spec(
                name="Different Alignments",
                channels=[1, 2],
                data_types=[sy.DataType.UINT8, sy.DataType.FLOAT32],
                frame=sy.Frame(
                    channels=[1, 2],
                    series=[
                        sy.Series(
                            data_type=sy.DataType.UINT8,
                            data=np.array([1], dtype=np.uint8),
                            alignment=5,
                        ),
                        sy.Series(
                            data_type=sy.DataType.FLOAT32,
                            data=np.array([1, 2, 3, 4], dtype=np.float32),
                            alignment=10,
                        ),
                    ],
                ),
            ),
            Spec(
                name="Regression 1",
                channels=[1, 2],
                data_types=[sy.DataType.TIMESTAMP, sy.DataType.FLOAT32],
                frame=sy.Frame(
                    channels=[2],
                    series=[
                        sy.Series(
                            data_type=sy.DataType.FLOAT32,
                            data=np.array([1, 2, 3, 4], dtype=np.float32),
                            time_range=sy.TimeRange(start=0, end=5),
                        ),
                    ],
                ),
            ),
            Spec(
                name="Variable Data Types",
                channels=[1, 2],
                data_types=[sy.DataType.UINT8, sy.DataType.STRING, sy.DataType.JSON],
                frame=sy.Frame(
                    channels=[1, 2],
                    series=[
                        sy.Series(
                            data_type=sy.DataType.UINT8,
                            data=np.array([1], dtype=np.uint8),
                            time_range=sy.TimeRange(start=0, end=5),
                        ),
                        sy.Series(["cat", "dog", "orange"]),
                        sy.Series(
                            [{"key": "value"}, {"key1": "value1"}],
                            data_type=sy.DataType.JSON,
                        ),
                    ],
                ),
            ),
        ],
    )
    def test_encoder_decoder(self, spec: Spec):
        codec = Codec(spec.channels, spec.data_types)
        encoded = codec.encode(spec.frame)
        decoded = codec.decode(encoded)
        assert len(decoded.keys) == len(spec.frame.channels)
        for i, key in enumerate(decoded.keys):
            dec_ser = decoded.series[i]
            or_ser = spec.frame[key].series[0]
            assert np.array_equal(list(dec_ser), list(or_ser))
            if or_ser.time_range is None:
                assert dec_ser.time_range == sy.TimeRange.ZERO
            else:
                assert dec_ser.time_range == or_ser.time_range
            assert dec_ser.alignment == or_ser.alignment

    def test_dynamic_codec_update(self):
        """Tests that the codec can be updated with new channels and correctly encode/decode after update"""
        codec = Codec()
        # First update and verification
        codec.update([1], [sy.DataType.INT32])
        frame = sy.Frame(
            channels=[1], series=[sy.Series(data=np.array([1, 2, 3], dtype=np.int32))]
        )
        encoded = codec.encode(frame)
        decoded = codec.decode(encoded)
        assert len(decoded.keys) == 1
        assert decoded.keys[0] == 1
        assert np.array_equal(list(decoded.series[0]), [1, 2, 3])

        # Second update and verification
        codec.update([2], [sy.DataType.INT64])
        frame2 = sy.Frame(
            channels=[2], series=[sy.Series(data=np.array([1, 2, 3], dtype=np.int64))]
        )
        encoded2 = codec.encode(frame2)
        decoded2 = codec.decode(encoded2)
        assert len(decoded2.keys) == 1
        assert decoded2.keys[0] == 2
        assert np.array_equal(list(decoded2.series[0]), [1, 2, 3])

    def test_uninitialized_codec(self):
        """Tests that using an uninitialized codec raises appropriate errors"""
        codec = Codec()
        frame = sy.Frame(
            channels=[1], series=[sy.Series(data=np.array([1, 2, 3], dtype=np.int32))]
        )
        with pytest.raises(ValueError):
            codec.encode(frame)

    def test_out_of_sync_codecs(self):
        """Tests correct behavior when encoder and decoder are out of sync with different states"""
        encoder = Codec()
        decoder = Codec()

        # Initial state - both in sync
        encoder.update([1], [sy.DataType.INT32])
        decoder.update([1], [sy.DataType.INT32])
        frame = sy.Frame(
            channels=[1], series=[sy.Series(data=np.array([1, 2, 3], dtype=np.int32))]
        )
        encoded = encoder.encode(frame)
        decoded = decoder.decode(encoded)
        assert len(decoded.keys) == 1
        assert decoded.keys[0] == 1
        assert np.array_equal(list(decoded.series[0]), [1, 2, 3])

        # Decoder updates but encoder doesn't - should still work with old format
        decoder.update([2], [sy.DataType.INT64])
        encoded = encoder.encode(frame)
        decoded = decoder.decode(encoded)
        assert len(decoded.keys) == 1
        assert decoded.keys[0] == 1
        assert np.array_equal(list(decoded.series[0]), [1, 2, 3])

        # Encoder updates - old frame should now fail
        encoder.update([2], [sy.DataType.INT64])
        with pytest.raises(sy.ValidationError):
            encoder.encode(frame)

        # New frame with updated channel should work
        frame2 = sy.Frame(
            channels=[2], series=[sy.Series(data=np.array([1, 2, 3], dtype=np.int64))]
        )
        encoded2 = encoder.encode(frame2)
        decoded2 = decoder.decode(encoded2)
        assert len(decoded2.keys) == 1
        assert decoded2.keys[0] == 2
        assert np.array_equal(list(decoded2.series[0]), [1, 2, 3])



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example demonstrates streaming data from a large number of channels into Synnax,
while also acknowledging commands sent to a large number of simulated valves. This
example serves as a basic simulated data acquisition system (DAQ). While this example is
running, you could open a schematic in the Synnax Console to control valves and view
sensors.
"""

import numpy as np

import synnax as sy

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

NUM_VALVES = 10
NUM_SENSORS = 10

# Some lists to store our channels.
valve_commands = list()
valve_responses = list()

# Maps the keys of valve command channels to response channels.
command_to_response = {}

# Stores the timestamps for both the sensors and the valve responses.
sensor_time_channel = client.channels.create(
    name="simulation_time",
    is_index=True,
    data_type=sy.DataType.TIMESTAMP,
    retrieve_if_name_exists=True,
)

# Create the necessary channels for each valve.
for i in range(NUM_VALVES):
    # The index channel for the command is used to track the time at which the command
    # was sent. We need to have separate indexes for each command channel so that these
    # channels can be written to independently.
    cmd_index_channel = client.channels.create(
        name=f"simulation_valve_command_{i}_time",
        is_index=True,
        data_type=sy.DataType.TIMESTAMP,
        retrieve_if_name_exists=True,
    )
    cmd_res = client.channels.create(
        [
            # The command channel is used to send a command to the valve.
            sy.Channel(
                name=f"simulation_valve_command_{i}",
                index=cmd_index_channel.key,
                data_type=sy.DataType.UINT8,
            ),
            # The response channel is used to acknowledge the response from our simulated
            # DAQ.
            sy.Channel(
                name=f"simulation_valve_response_{i}",
                index=sensor_time_channel.key,
                data_type=sy.DataType.UINT8,
            ),
        ],
        retrieve_if_name_exists=True,
    )
    cmd = cmd_res[0]
    res = cmd_res[1]
    valve_commands.append(cmd)
    valve_responses.append(res)
    command_to_response[cmd.key] = res

# Defining the sensor channels to create.
sensors = [
    sy.Channel(
        name=f"simulation_sensor_{i}",
        index=sensor_time_channel.key,
        data_type=sy.DataType.FLOAT32,
    )
    for i in range(NUM_SENSORS)
]

# Actually creating the sensor channels on the Synnax cluster.
sensors = client.channels.create(sensors, retrieve_if_name_exists=True)

# Define the list of channels we'll write to (the sensor channels and the valve
# responses)
write_to = [
    sensor_time_channel.key,  # sensor time
    *[s.key for s in sensors],  # sensor data
    *[v.key for v in valve_responses],  # valve responses
]

# Define the list of channels we'll read from (the incoming valve commands)
read_from = [v.key for v in valve_commands]

# Define a rate at which we'll write data.
loop = sy.Loop(sy.Rate.HZ * 100)

# Set up the initial state of the valves to 0 (closed).
sensor_states = {v.key: np.uint8(False) for v in valve_responses}

# Open a streamer to listen for incoming valve commands.
with client.open_streamer([channel.key for channel in valve_commands]) as streamer:
    i = 0
    # Open a writer to write data to Synnax.
    with client.open_writer(sy.TimeStamp.now(), write_to) as writer:
        start = sy.TimeStamp.now()
        while loop.wait():
            # If we've received a command, update the state of the corresponding valve.
            frame = streamer.read(timeout=0)
            if frame is not None:
                for channel in frame.channels:
                    # 1 is open, 0 is closed
                    sensor_states[command_to_response[channel].key] = np.uint8(
                        frame[channel][-1] > 0.9
                    )
            for j, channel in enumerate(sensors):
                sensor_states[channel.key] = np.float32(np.sin(i / 1000) + j / 100)
            sensor_states[sensor_time_channel.key] = sy.TimeStamp.now()
            writer.write(sensor_states)
            i += 1



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import random

import synnax as sy
from examples.simulators.simdaq import SimDAQ


class PressSimDAQ(SimDAQ):
    """Simulates pressurization system with valves and pressure sensor."""

    description = "Run press simulator standalone"
    end_cmd_channel = "end_test_cmd"

    def _create_channels(self) -> None:
        self.log("Creating channels...")
        client = self.client

        self.daq_time_ch = client.channels.create(
            name="daq_time", is_index=True, retrieve_if_name_exists=True
        )

        self.press_pt = client.channels.create(
            name="press_pt",
            index=self.daq_time_ch.key,
            data_type=sy.DataType.FLOAT32,
            retrieve_if_name_exists=True,
        )

        self.test_flag_cmd = client.channels.create(
            name="test_flag_cmd",
            data_type=sy.DataType.UINT8,
            retrieve_if_name_exists=True,
            virtual=True,
        )

        self.end_test_cmd = client.channels.create(
            name="end_test_cmd",
            data_type=sy.DataType.UINT8,
            retrieve_if_name_exists=True,
            virtual=True,
        )

        client.write(sy.TimeStamp.now(), self.test_flag_cmd.key, [0])
        client.write(sy.TimeStamp.now(), self.end_test_cmd.key, [0])

        self.press_vlv_cmd_time = client.channels.create(
            name="press_vlv_cmd_time",
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
            retrieve_if_name_exists=True,
        )

        self.press_vlv_cmd = client.channels.create(
            name="press_vlv_cmd",
            data_type=sy.DataType.UINT8,
            retrieve_if_name_exists=True,
            index=self.press_vlv_cmd_time.key,
        )

        self.press_vlv_state = client.channels.create(
            name="press_vlv_state",
            index=self.daq_time_ch.key,
            data_type=sy.DataType.UINT8,
            retrieve_if_name_exists=True,
        )

        self.vent_vlv_cmd_time = client.channels.create(
            name="vent_vlv_cmd_time",
            is_index=True,
            data_type=sy.DataType.TIMESTAMP,
            retrieve_if_name_exists=True,
        )

        self.vent_vlv_cmd = client.channels.create(
            name="vent_vlv_cmd",
            data_type=sy.DataType.UINT8,
            retrieve_if_name_exists=True,
            index=self.vent_vlv_cmd_time.key,
        )

        self.vent_vlv_state = client.channels.create(
            name="vent_vlv_state",
            index=self.daq_time_ch.key,
            data_type=sy.DataType.UINT8,
            retrieve_if_name_exists=True,
        )

        self.press_setpoint_cmd = client.channels.create(
            name="press_setpoint_cmd",
            data_type=sy.DataType.FLOAT32,
            retrieve_if_name_exists=True,
            virtual=True,
        )

        client.write(sy.TimeStamp.now(), self.press_setpoint_cmd.key, [0.0])

        now = sy.TimeStamp.now()
        client.write(
            now,
            {
                self.daq_time_ch.key: [now],
                self.press_vlv_state.key: [0],
                self.vent_vlv_state.key: [0],
                self.press_pt.key: [0.0],
            },
        )
        self.log("Channels created successfully")

    def _run_loop(self) -> None:
        self.log("Starting simulation loop...")
        loop = sy.Loop(sy.Rate.HZ * 100)
        loop_count = 0

        state = {
            "daq_time": sy.TimeStamp.now(),
            "press_vlv_state": 0,
            "vent_vlv_state": 0,
            "press_pt": 0.0,
        }

        with self.client.open_streamer(["press_vlv_cmd", "vent_vlv_cmd"]) as streamer:
            with self.client.open_writer(
                start=sy.TimeStamp.now(),
                channels=[
                    self.daq_time_ch.name,
                    "press_vlv_state",
                    "vent_vlv_state",
                    "press_pt",
                ],
                name="Simulated DAQ",
            ) as writer:
                while self._running and loop.wait():
                    while True:
                        frame = streamer.read(timeout=0)
                        if frame is not None:
                            vent_vlv_cmd = frame.get("vent_vlv_cmd")
                            if len(vent_vlv_cmd) > 0:
                                new_val = vent_vlv_cmd[-1]
                                if hasattr(new_val, "item"):
                                    new_val = new_val.item()
                                if new_val != state["vent_vlv_state"]:
                                    self.log(
                                        f"Vent valve: {state['vent_vlv_state']} -> {new_val}"
                                    )
                                state["vent_vlv_state"] = new_val

                            press_vlv_cmd = frame.get("press_vlv_cmd")
                            if len(press_vlv_cmd) > 0:
                                new_val = press_vlv_cmd[-1]
                                if hasattr(new_val, "item"):
                                    new_val = new_val.item()
                                if new_val != state["press_vlv_state"]:
                                    self.log(
                                        f"Press valve: {state['press_vlv_state']} -> {new_val}"
                                    )
                                state["press_vlv_state"] = new_val
                        else:
                            break

                    if state["press_vlv_state"] == 1:
                        state["press_pt"] += 0.2

                    if state["vent_vlv_state"] == 1:
                        state["press_pt"] -= 0.2

                    if state["press_pt"] < 0:
                        state["press_pt"] = 0

                    state["press_pt"] += random.uniform(-0.05, 0.05)
                    state["daq_time"] = sy.TimeStamp.now()
                    writer.write(state)

                    loop_count += 1
                    if loop_count % 100 == 0:
                        self.log(
                            f"press_pt={state['press_pt']:.2f}, "
                            f"press_vlv={state['press_vlv_state']}, "
                            f"vent_vlv={state['vent_vlv_state']}"
                        )

        self.log("Simulation loop stopped")


if __name__ == "__main__":
    PressSimDAQ.main()



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import asyncio
import multiprocessing
import os
import signal
import socket
import sys
from abc import abstractmethod
from multiprocessing.process import BaseProcess

import synnax as sy
from examples.simulators.simulator import Simulator


class DeviceSim(Simulator):
    """Base class for protocol device simulators.

    These simulators expose network endpoints (Modbus TCP, OPC UA, etc.)
    but do NOT interact with Synnax. They run async servers in a separate
    process to avoid event loop conflicts.

    Subclasses must implement:
    - _run_server(): async coroutine that starts and runs the server

    Class attributes:
    - host: network host the server binds to
    - port: TCP port the server listens on (used for readiness check)
    - startup_timeout: max time to wait for the server to accept connections
    - device_name: name used for Synnax device registration
    """

    host: str
    port: int
    startup_timeout: sy.TimeSpan = 10 * sy.TimeSpan.SECOND
    device_name: str

    def __init__(self, rate: sy.Rate = 50 * sy.Rate.HZ, verbose: bool = False):
        super().__init__(verbose=verbose)
        self.rate = rate
        self.process: BaseProcess | None = None

    def _kill_port_occupant(self) -> None:
        """Kill any process currently listening on self.port."""
        try:
            with socket.create_connection((self.host, self.port), timeout=0.5):
                pass
        except OSError:
            return  # Port is free
        self.log(f"Port {self.port} is in use, killing stale process")
        killed = False
        try:
            import psutil

            # net_connections queries all TCP sockets in one syscall — much
            # faster than iterating every process with process_iter().
            for conn in psutil.net_connections(kind="tcp"):
                if conn.status == "LISTEN" and conn.laddr.port == self.port:
                    if conn.pid is None:
                        continue
                    try:
                        psutil.Process(conn.pid).kill()
                        self.log(f"Killed stale PID {conn.pid} on port {self.port}")
                        killed = True
                    except (psutil.NoSuchProcess, psutil.AccessDenied):
                        pass
        except Exception as e:
            self.log(f"Could not kill port occupant: {e}")
        if killed:
            sy.sleep(1)  # Give OS time to release the port

    def start(self) -> None:
        """Start the device server in a subprocess.

        Blocks until the server is accepting TCP connections on its port,
        or raises RuntimeError if the process dies or the timeout expires.
        """
        self._kill_port_occupant()
        self.process = multiprocessing.Process(
            target=self._subprocess_entry, daemon=True
        )
        self.process.start()
        self._running = True
        self.log(f"Server started with PID: {self.process.pid}")
        self._wait_for_ready()

    def _wait_for_ready(self) -> None:
        """Poll until the server is accepting connections on its port."""
        timeout = self.startup_timeout
        timer = sy.Timer()
        while timer.elapsed() < timeout:
            if not self.process.is_alive():
                raise RuntimeError(
                    f"Server process died during startup "
                    f"(exit code: {self.process.exitcode})"
                )
            try:
                with socket.create_connection((self.host, self.port), timeout=0.5):
                    self.log(f"Server ready on {self.host}:{self.port}")
                    return
            except OSError:
                sy.sleep(0.1)
        raise RuntimeError(
            f"Server not ready on {self.host}:{self.port} "
            f"after {self.startup_timeout}"
        )

    def stop(self, timeout: sy.TimeSpan = 5 * sy.TimeSpan.SECOND) -> None:
        """Terminate the server subprocess."""
        self._running = False
        if self.process is None:
            return
        if not self.process.is_alive():
            self.process = None
            return
        timeout_secs = float(timeout / sy.TimeSpan.SECOND)
        try:
            self.process.terminate()
            self.process.join(timeout=timeout_secs)
            if self.process.is_alive():
                self.process.kill()
                self.process.join(timeout=2)
        except Exception as e:
            self.log(f"Error terminating server: {e}")
        finally:
            self.process = None
            sy.sleep(1)

    def _subprocess_entry(self) -> None:
        """Entry point for the subprocess."""
        devnull_fd = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull_fd, 1)
        os.dup2(devnull_fd, 2)
        os.close(devnull_fd)
        sys.stdout = open(os.devnull, "w")
        sys.stderr = open(os.devnull, "w")
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        if sys.platform != "win32":
            signal.signal(signal.SIGTERM, signal.SIG_DFL)
        asyncio.run(self._run_server())

    @staticmethod
    def create_device(rack_key: int) -> sy.device.Device:
        raise NotImplementedError

    @abstractmethod
    async def _run_server(self) -> None:
        """The async server coroutine. Must run indefinitely until killed."""
        ...



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from collections.abc import Iterator
from math import ceil
from pathlib import Path

import pandas as pd
from nptdms import TdmsFile

from synnax.io.matcher import new_extension_matcher
from synnax.io.protocol import (
    ChannelMeta,
    ColumnFileReader,
    ReaderType,
)

TDMSMatcher = new_extension_matcher(["tdms"])


class TDMSReader(TDMSMatcher):  # type: ignore
    """A ColReader implementation for TDMS files."""

    channel_keys: set[str]
    chunk_size: int

    _current_chunk: int
    _n_chunks: int | None

    _path: Path
    _channels: list[ChannelMeta] | None

    # Doing a protocol implementation check here because
    # it's hard for pyright to handle factories that return
    # protocol classes.
    def _(self) -> ColumnFileReader:
        return self

    def __init__(
        self,
        path: Path,
        keys: list[str] | None = None,
        chunk_size: int | None = None,
    ):
        self._path = path
        self.chunk_size = chunk_size or int(1e5)

        self._channels = None
        self._current_chunk = 0
        self._n_chunks = None
        if keys is not None:
            self.channel_keys = set(keys)
        else:
            self.set_keys_from_file()

    def channels(self) -> list[ChannelMeta]:
        """:returns : a list of channel metadata for the file."""
        if self._channels is not None:
            return self._channels

        self._channels: list[ChannelMeta] = list()
        with TdmsFile.open(self._path) as tdms_file:
            for group in tdms_file.groups():
                for channel in group.channels():
                    self._channels.append(
                        ChannelMeta(name=channel.name, meta_data=dict())
                    )

        return self._channels

    def set_chunk_size(self, chunk_size: int) -> None:
        self.chunk_size = chunk_size

    def set_keys(self, keys: list[str]) -> None:
        self.channel_keys = set(keys)

    def set_keys_from_file(self) -> set[str]:
        self.channel_keys = set([ch.name for ch in self.channels()])
        return self.channel_keys

    @classmethod
    def type(cls) -> ReaderType:
        """:returns : the type of reader."""
        return ReaderType.Column

    def path(self) -> Path:
        """:returns: the path to the file."""
        return self._path

    def nsamples(self) -> int:
        """:returns: the number of samples in the file."""
        return self.chunk_size * self.n_chunks * len(self.channels())

    def seek_first(self) -> None:
        """Seeks the reader to the first sample in the file."""
        self._current_chunk = 0

    def read(self, *keys: str) -> pd.DataFrame:
        """Reads a dataframe with chunk size * number of columns samples. The returned
        dataframe contains columns for each key in keys.
        """
        # if we already read everything, return an empty chunk
        if self._current_chunk >= self.n_chunks:
            return pd.DataFrame()

        _keys: set[str] = self.channel_keys if (len(keys) == 0) else set(keys)

        # https://nptdms.readthedocs.io/en/stable/reading.html
        # https://github.com/adamreeve/npTDMS/issues/263
        # https://nptdms.readthedocs.io/en/stable/reading.html
        data = dict()
        with TdmsFile.open(self._path) as tdms_file:
            for group in tdms_file.groups():
                for channel in group.channels():
                    if channel.name in _keys:
                        data[channel.name] = channel[
                            self._current_chunk
                            * self.chunk_size : (self._current_chunk + 1)
                            * self.chunk_size
                        ]
            self._current_chunk += 1

        return pd.DataFrame(data)

    def __iter__(self) -> Iterator[pd.DataFrame]:
        return (self.read() for _ in range(self.n_chunks))

    @property
    def n_chunks(self) -> int:
        """Returns number of chunks in a channel.
        Assumes every channel has the same number of chunks
        """
        if self._n_chunks is None:
            # assume every channel has the same number of data
            with TdmsFile.open(self._path) as tdms_file:
                group0 = tdms_file.groups()[0]
                channel0 = group0.channels()[0]
                self._n_chunks = ceil(len(channel0) / self.chunk_size)

        assert self._n_chunks is not None
        return self._n_chunks



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This script connects a LabJack device by registering it in Synnax.

If the device is already connected, the script exits successfully.
Otherwise, it prompts to create a new LabJack device connection automatically.

Before running this example:
1. Start the Synnax Driver (if not already running).

2. Connect your LabJack device via USB, Ethernet, or WiFi.

3. Login to Synnax (if not already logged in):
   uv run sy login

4. Run this script:
   uv run python examples/labjack/connect_device.py

Configuration:
    Modify the constants below to match your LabJack device configuration.
"""

import synnax as sy

# Configuration
DEVICE_NAME = "My LabJack T7"
MODEL = sy.labjack.T7  # Options: T4, T7, T8
IDENTIFIER = "ANY"  # Options: serial number, IP address, device name, or "ANY"
CONNECTION_TYPE = "ANY"  # Options: "ANY", "USB", "TCP", "ETHERNET", "WIFI"

# Connect to Synnax
client = sy.Synnax()

print("=" * 70)
print("LabJack Device Connection Script")
print("=" * 70)
print(f"Target Device Name: {DEVICE_NAME}")
print(f"Model: {MODEL}")
print(f"Identifier: {IDENTIFIER}")
print(f"Connection Type: {CONNECTION_TYPE}")
print()

# Check if device already exists
existing_device = client.devices.retrieve(name=DEVICE_NAME, ignore_not_found=True)

if existing_device is not None:
    print("✓ Device already connected!")
    print(f"- Name: {existing_device.name}")
    print(f"- Model: {existing_device.model}")
    print(f"- Key: {existing_device.key}")
    print(f"- Location: {existing_device.location}")
    print()
    print("No action needed - device is ready to use.")
    print("=" * 70)
    exit(0)

# Device not found - prompt user for connection method
print("Device not found")
print()
print("Would you like to connect automatically? [Y/N]: ", end="", flush=True)

try:
    response = input().strip().lower()
except (KeyboardInterrupt, EOFError):
    print("\n")
    response = "n"

if response in ("", "y", "yes"):
    print()
    print("Connecting to LabJack device...")

    try:
        # Get the embedded rack (local driver rack)
        rack = client.racks.retrieve_embedded_rack()
        print(f"Using rack: {rack.name} (key={rack.key})")

        # Determine location string based on connection type
        if IDENTIFIER == "ANY":
            location = CONNECTION_TYPE
        else:
            location = IDENTIFIER

        device = sy.labjack.Device(
            model=MODEL,
            identifier=IDENTIFIER,
            name=DEVICE_NAME,
            location=location,
            rack=rack.key,
            connection_type=CONNECTION_TYPE,
        )

        created_device = client.devices.create(device)

        print("✓ Device connected successfully!")
        print(f"  - Name: {created_device.name}")
        print(f"  - Model: {created_device.model}")
        print(f"  - Key: {created_device.key}")
        print(f"  - Location: {created_device.location}")
        print(f"  - Rack: {rack.name}")
        print()
        print("Device is ready to use.")
        print()
        print("Next steps:")
        print("  - Run examples/labjack/read_task.py to read data")
        print("  - Run examples/labjack/write_task.py to control outputs")
        print("=" * 70)
        exit(0)

    except Exception as e:
        print(f"✗ Failed to connect device: {e}")
        print()
        print("Please try connecting manually (see instructions below).")
        response = "n"  # Fall through to manual instructions

if response in ("n", "no") or response not in ("", "y", "yes"):
    # Manual connection instructions
    print()
    print("To connect manually:")
    print("1. Open Synnax Console")
    print("2. Go to Resources → Devices")
    print("3. Add a new LabJack device:")
    print(f"   - Name: {DEVICE_NAME}")
    print(f"   - Model: {MODEL}")
    print(f"   - Identifier: {IDENTIFIER}")
    print(f"   - Connection Type: {CONNECTION_TYPE}")
    print()
    print("Identifier options:")
    print("  - 'ANY': Connect to any available device")
    print("  - Serial number (e.g., '470012345')")
    print("  - IP address (e.g., '192.168.1.100')")
    print("  - Device name (e.g., 'MyLabJack')")
    print()
    print("Connection Type options:")
    print("  - 'ANY': Use any available connection")
    print("  - 'USB': Connect via USB only")
    print("  - 'TCP' or 'ETHERNET': Connect via Ethernet")
    print("  - 'WIFI': Connect via WiFi")
    print("=" * 70)
    exit(1)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import keyring
from pydantic import BaseModel

from synnax.config.file import ConfigFile
from synnax.options import Options


class ClusterConfig(BaseModel):
    options: Options


class ClustersConfig:
    internal: ConfigFile

    def __init__(self, config: ConfigFile):
        self.internal = config

    def get(self, key: str = "default") -> ClusterConfig | None:
        c = self.internal.get(f"clusters.{key}")
        if c is None:
            return None
        opts = c["options"]
        pwd = keyring.get_password("synnax", key)
        pwd = pwd or ""
        return ClusterConfig(options=Options(**opts, password=pwd))

    def set(self, c: ClusterConfig, key: str = "default") -> None:
        p = c.model_dump()
        keyring.set_password("synnax", key, p["options"].pop("password"))
        self.internal.set(f"clusters.{key}", p)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from typing import overload
from uuid import UUID

from freighter import Empty, UnaryClient, send_required
from pydantic import BaseModel, PrivateAttr

from synnax.arc.payload import (
    Graph,
    Key,
    Mode,
    Payload,
    Text,
    ontology_id,
)
from synnax.exceptions import MultipleFoundError, NotFoundError
from synnax.ontology.payload import ID
from synnax.util.normalize import normalize


class _CreateRequest(BaseModel):
    arcs: list[Payload] = []


class _CreateResponse(BaseModel):
    arcs: list[Payload] = []


class _RetrieveRequest(BaseModel):
    keys: list[UUID] | None = None
    names: list[str] | None = None
    search_term: str | None = None
    limit: int | None = None
    offset: int | None = None


class _RetrieveResponse(BaseModel):
    arcs: list[Payload] | None = None


class _DeleteRequest(BaseModel):
    keys: list[UUID]


class Arc(Payload):
    """An Arc is a user-defined automation sequence or control program.

    Arcs can be created in two modes:
    - 'text': Source code written in the Arc programming language
    - 'graph': Visual programming using a node-based editor
    """

    __client: Client | None = PrivateAttr(None)

    def __init__(
        self,
        *,
        key: UUID = UUID(int=0),
        name: str = "",
        graph: Graph | None = None,
        text: Text | None = None,
        version: str = "",
        mode: Mode = "text",
        _client: Client | None = None,
    ):
        super().__init__(
            key=key,
            name=name,
            graph=graph if graph is not None else Graph(),
            text=text if text is not None else Text(),
            version=version,
            mode=mode,
        )
        self.__client = _client

    @property
    def ontology_id(self) -> ID:
        return ontology_id(self.key)

    def to_payload(self) -> Payload:
        return Payload(
            key=self.key,
            name=self.name,
            graph=self.graph,
            text=self.text,
            version=self.version,
            mode=self.mode,
        )


class Client:
    """Client for managing Arc programs in the Synnax cluster."""

    _client: UnaryClient

    def __init__(self, client: UnaryClient) -> None:
        self._client = client

    @overload
    def create(
        self,
        *,
        name: str,
        graph: Graph | None = None,
        text: Text | None = None,
        version: str = "",
        mode: Mode = "text",
    ) -> Arc: ...

    @overload
    def create(self, arc: Arc) -> Arc: ...

    @overload
    def create(self, arc: list[Arc]) -> list[Arc]: ...

    def create(
        self,
        arc: Arc | list[Arc] | None = None,
        *,
        name: str = "",
        graph: Graph | None = None,
        text: Text | None = None,
        version: str = "",
        mode: Mode = "text",
    ) -> Arc | list[Arc]:
        is_single = True
        if arc is not None:
            if isinstance(arc, list):
                is_single = False
                to_create = [a.to_payload() for a in arc]
            else:
                to_create = [arc.to_payload()]
        else:
            to_create = [
                Payload(
                    name=name,
                    graph=graph if graph is not None else Graph(),
                    text=text if text is not None else Text(),
                    version=version,
                    mode=mode,
                )
            ]

        res = send_required(
            self._client,
            "/arc/create",
            _CreateRequest(arcs=to_create),
            _CreateResponse,
        ).arcs
        created = self.__sugar(res)
        return created[0] if is_single else created

    @overload
    def retrieve(self, *, key: Key) -> Arc: ...

    @overload
    def retrieve(self, *, name: str) -> Arc: ...

    @overload
    def retrieve(
        self,
        *,
        keys: list[Key] | None = None,
        names: list[str] | None = None,
        search_term: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[Arc]: ...

    def retrieve(
        self,
        *,
        key: Key | None = None,
        name: str | None = None,
        keys: list[Key] | None = None,
        names: list[str] | None = None,
        search_term: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> Arc | list[Arc]:
        if key is not None:
            keys = [key]
        if name is not None:
            names = [name]

        is_single = key is not None or name is not None
        res = send_required(
            self._client,
            "/arc/retrieve",
            _RetrieveRequest(
                keys=keys,
                names=names,
                search_term=search_term,
                limit=limit,
                offset=offset,
            ),
            _RetrieveResponse,
        )

        arcs = self.__sugar(res.arcs or [])
        if not is_single:
            return arcs
        if len(arcs) == 0:
            identifier = key if key is not None else name
            raise NotFoundError(f"Arc {identifier} not found")
        if len(arcs) > 1:
            identifier = key if key is not None else name
            raise MultipleFoundError(f"Multiple Arcs matching {identifier} found")
        return arcs[0]

    def delete(self, keys: Key | list[Key]) -> None:
        send_required(
            self._client,
            "/arc/delete",
            _DeleteRequest(keys=normalize(keys)),
            Empty,
        )

    def __sugar(self, payloads: list[Payload]) -> list[Arc]:
        return [
            Arc(
                key=p.key,
                name=p.name,
                graph=p.graph,
                text=p.text,
                version=p.version,
                mode=p.mode,
                _client=self,
            )
            for p in payloads
        ]



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""EtherCAT driver types for Synnax Python SDK.

This module provides Python types for configuring EtherCAT read and write tasks
that communicate with EtherCAT slave devices via the Synnax driver.

EtherCAT channels can be configured in two modes:
- **Automatic**: Reference PDOs by name (e.g., "Position actual value"). The driver
  looks up the PDO address from the slave device properties discovered during scanning.
- **Manual**: Specify PDO addresses directly (index, subindex, bit_length, data_type).

Example:
    >>> import synnax as sy
    >>> from synnax import ethercat
    >>>
    >>> # Create an automatic input channel (PDO resolved by name)
    >>> position_ch = ethercat.AutomaticInputChan(
    ...     device="slave-device-key",
    ...     pdo="Position actual value",
    ...     channel=position_channel.key,
    ... )
    >>>
    >>> # Create a read task
    >>> read_task = ethercat.ReadTask(
    ...     name="EtherCAT Read",
    ...     sample_rate=1000,
    ...     stream_rate=100,
    ...     channels=[position_ch],
    ... )
    >>> client.tasks.configure(read_task)
"""

from typing import Any, Literal
from uuid import uuid4

from pydantic import BaseModel, Field, field_validator

from synnax import channel, device, task
from synnax.telem import CrudeRate

# Device identifiers - must match driver expectations
MAKE = "EtherCAT"
MODEL = "Slave"


class PDOEntry(BaseModel):
    """Information about a single PDO entry from device discovery.

    This model represents a PDO (Process Data Object) entry stored in
    slave device properties after scanning. Note that this represents an
    individual entry within a PDO, not the PDO container itself.

    :param name: Human-readable name of the PDO entry (e.g., "Position actual value").
    :param pdo_index: Parent PDO index (e.g., 0x1A00 for TxPDO, 0x1600 for RxPDO).
    :param index: CoE object dictionary index (e.g., 0x6064 = 24676).
    :param sub_index: CoE object dictionary subindex.
    :param bit_length: Size of the data in bits.
    :param data_type: Data type string (e.g., "uint16", "int32", "float32").
    """

    name: str
    pdo_index: int = Field(default=0, ge=0, le=65535)
    index: int = Field(ge=0, le=65535)
    sub_index: int = Field(ge=0, le=255)
    bit_length: int = Field(ge=1, le=64)
    data_type: str


class BaseChan(BaseModel):
    """Base class for all EtherCAT channels."""

    key: str = ""
    "A unique key to identify this channel configuration."
    enabled: bool = True
    "Whether the channel is enabled for data exchange."
    device: str = Field(min_length=1)
    "The key of the Synnax slave device this channel belongs to."

    def __init__(self, **data: Any) -> None:
        if "key" not in data or not data["key"]:
            data["key"] = str(uuid4())
        super().__init__(**data)


class AutomaticInputChan(BaseChan):
    """Input channel that resolves PDO address from slave device properties by name.

    Automatic channels reference PDOs by their human-readable name. The driver
    looks up the PDO address (index, subindex, bit_length, data_type) from the
    slave device properties that were populated during device scanning.

    This is the recommended approach as it's more readable and less error-prone
    than specifying raw PDO addresses.

    Example:
        >>> # Read position from a servo drive
        >>> position_ch = AutomaticInputChan(
        ...     device="servo-drive-key",
        ...     pdo="Position actual value",
        ...     channel=position_channel.key,
        ... )

    :param device: The key of the Synnax slave device this channel belongs to.
    :param pdo: The name of the PDO to read (must match a name in device properties).
    :param channel: The Synnax channel key that data will be written to.
    :param enabled: Whether the channel is enabled for data acquisition.
    :param key: Unique identifier (auto-generated if not provided).
    """

    type: Literal["automatic"] = "automatic"
    pdo: str = Field(min_length=1)
    "The name of the PDO to look up in slave device properties."
    channel: channel.Key
    "The Synnax channel key that will be written to during acquisition."


class ManualInputChan(BaseChan):
    """Input channel with PDO address specified directly.

    Manual channels allow specifying the exact PDO address when the PDO name
    is not known or when working with devices that don't have properly
    populated PDO names.

    Example:
        >>> # Read a PDO at index 0x6064, sub_index 0
        >>> position_ch = ManualInputChan(
        ...     device="servo-drive-key",
        ...     index=0x6064,
        ...     sub_index=0,
        ...     bit_length=32,
        ...     data_type="int32",
        ...     channel=position_channel.key,
        ... )

    :param device: The key of the Synnax slave device this channel belongs to.
    :param index: CoE object dictionary index (0-65535).
    :param sub_index: CoE object dictionary subindex (0-255).
    :param bit_length: Size of the data in bits (1-64).
    :param data_type: Data type string (e.g., "uint16", "int32", "float32").
    :param channel: The Synnax channel key that data will be written to.
    :param enabled: Whether the channel is enabled for data acquisition.
    :param key: Unique identifier (auto-generated if not provided).
    """

    type: Literal["manual"] = "manual"
    index: int = Field(ge=0, le=65535)
    "CoE object dictionary index (e.g., 0x6064 = 24676)."
    sub_index: int = Field(ge=0, le=255)
    "CoE object dictionary subindex."
    bit_length: int = Field(ge=1, le=64)
    "Size of the data in bits."
    data_type: str
    "Data type string (e.g., 'uint8', 'int16', 'uint32', 'int32', 'float32')."
    channel: channel.Key
    "The Synnax channel key that will be written to during acquisition."


# Union type for all input channels
InputChan = AutomaticInputChan | ManualInputChan


class AutomaticOutputChan(BaseChan):
    """Output channel that resolves PDO address from slave device properties by name.

    Automatic output channels reference PDOs by their human-readable name. The driver
    looks up the PDO address from the slave device properties.

    Example:
        >>> # Control velocity on a servo drive
        >>> velocity_cmd = AutomaticOutputChan(
        ...     device="servo-drive-key",
        ...     pdo="Target velocity",
        ...     cmd_channel=velocity_command.key,
        ...     state_channel=velocity_state.key,  # Optional state feedback
        ... )

    :param device: The key of the Synnax slave device this channel belongs to.
    :param pdo: The name of the PDO to write (must match a name in device properties).
    :param cmd_channel: The Synnax channel key to receive command values from.
    :param state_channel: The Synnax channel key to write state feedback to (optional).
    :param enabled: Whether the channel is enabled for output operations.
    :param key: Unique identifier (auto-generated if not provided).
    """

    type: Literal["automatic"] = "automatic"
    pdo: str = Field(min_length=1)
    "The name of the PDO to look up in slave device properties."
    cmd_channel: channel.Key
    "The Synnax channel key to receive command values from."
    state_channel: channel.Key = 0
    "The Synnax channel key to write state feedback to (0 to disable)."


class ManualOutputChan(BaseChan):
    """Output channel with PDO address specified directly.

    Manual output channels allow specifying the exact PDO address when the PDO name
    is not known or when working with devices that don't have properly populated
    PDO names.

    Example:
        >>> # Write to a PDO at index 0x60FF, sub_index 0
        >>> velocity_cmd = ManualOutputChan(
        ...     device="servo-drive-key",
        ...     index=0x60FF,
        ...     sub_index=0,
        ...     bit_length=32,
        ...     data_type="int32",
        ...     cmd_channel=velocity_command.key,
        ... )

    :param device: The key of the Synnax slave device this channel belongs to.
    :param index: CoE object dictionary index (0-65535).
    :param sub_index: CoE object dictionary subindex (0-255).
    :param bit_length: Size of the data in bits (1-64).
    :param data_type: Data type string (e.g., "uint16", "int32", "float32").
    :param cmd_channel: The Synnax channel key to receive command values from.
    :param state_channel: The Synnax channel key to write state feedback to (optional).
    :param enabled: Whether the channel is enabled for output operations.
    :param key: Unique identifier (auto-generated if not provided).
    """

    type: Literal["manual"] = "manual"
    index: int = Field(ge=0, le=65535)
    "CoE object dictionary index (e.g., 0x60FF = 24831)."
    sub_index: int = Field(ge=0, le=255)
    "CoE object dictionary subindex."
    bit_length: int = Field(ge=1, le=64)
    "Size of the data in bits."
    data_type: str
    "Data type string (e.g., 'uint8', 'int16', 'uint32', 'int32', 'float32')."
    cmd_channel: channel.Key
    "The Synnax channel key to receive command values from."
    state_channel: channel.Key = 0
    "The Synnax channel key to write state feedback to (0 to disable)."


# Union type for all output channels
OutputChan = AutomaticOutputChan | ManualOutputChan


class ReadTaskConfig(task.BaseReadConfig):
    """Configuration for an EtherCAT read task.

    Inherits common read task fields (sample_rate, stream_rate, data_saving,
    auto_start) from BaseReadTaskConfig and adds EtherCAT-specific channel
    configuration.

    The sample_rate must be divisible by stream_rate. For example, a sample_rate
    of 1000 Hz and stream_rate of 100 Hz will result in batches of 10 samples
    being streamed to Synnax every 10ms.

    EtherCAT supports high-frequency sampling. The default rate limits are set
    to 50 kHz, but actual achievable rates depend on the number of slaves and
    PDOs configured.

    :param sample_rate: The rate at which to sample data from EtherCAT slaves (Hz).
    :param stream_rate: The rate at which data is streamed to Synnax (Hz).
    :param data_saving: Whether to persist data to disk (True) or only stream (False).
    :param auto_start: Whether to start the task automatically when configured.
    :param channels: List of input channel configurations (automatic or manual).
    """

    channels: list[InputChan]
    "A list of input channel configurations to acquire data from."

    @field_validator("channels")
    def validate_channels_not_empty(cls, v: list[InputChan]) -> list[InputChan]:
        """Validate that at least one channel is provided."""
        if len(v) == 0:
            raise ValueError("Task must have at least one channel")
        return v


class WriteTaskConfig(task.BaseWriteConfig):
    """Configuration for an EtherCAT write task.

    Inherits common write task fields (device, auto_start) from BaseWriteTaskConfig
    and adds EtherCAT-specific fields for state feedback and execution rate.

    Note: The `device` field from BaseWriteTaskConfig is not used for EtherCAT
    write tasks since channels can span multiple slave devices. Each channel
    specifies its own device.

    :param state_rate: Rate at which state feedback is written to Synnax (Hz).
    :param execution_rate: Rate at which commands are executed on the bus (Hz).
    :param data_saving: Whether to persist state data to disk.
    :param auto_start: Whether to start the task automatically when configured.
    :param channels: List of output channel configurations (automatic or manual).
    """

    device: str = ""
    "Not used for EtherCAT - each channel specifies its own device."
    state_rate: float = Field(default=1.0, ge=0, le=10000)
    "Rate at which state feedback is written to Synnax (Hz)."
    execution_rate: float = Field(default=1000.0, ge=0, le=50000)
    "Rate at which commands are executed on the EtherCAT bus (Hz)."
    data_saving: bool = True
    "Whether to persist state data to disk."
    channels: list[OutputChan]
    "A list of output channel configurations to write to."

    @field_validator("channels")
    def validate_channels_not_empty(cls, v: list[OutputChan]) -> list[OutputChan]:
        """Validate that at least one channel is provided."""
        if len(v) == 0:
            raise ValueError("Task must have at least one channel")
        return v


class ReadTask(task.StarterStopperMixin, task.JSONConfigMixin, task.Protocol):
    """A read task for sampling data from EtherCAT slave devices.

    This task configures the Synnax driver to perform cyclic PDO exchange with
    EtherCAT slaves and write the acquired data to Synnax channels.

    The task supports both automatic channels (PDO resolved by name) and manual
    channels (PDO address specified directly).

    Example:
        >>> import synnax as sy
        >>> from synnax import ethercat
        >>>
        >>> # Create channels for time and position
        >>> time_ch = client.channels.create(
        ...     name="ecat_time",
        ...     data_type=sy.DataType.TIMESTAMP,
        ...     is_index=True,
        ... )
        >>> position_ch = client.channels.create(
        ...     name="position",
        ...     data_type=sy.DataType.INT32,
        ...     index=time_ch.key,
        ... )
        >>>
        >>> # Create the read task
        >>> read_task = ethercat.ReadTask(
        ...     name="EtherCAT Position Read",
        ...     sample_rate=1000,
        ...     stream_rate=100,
        ...     data_saving=True,
        ...     channels=[
        ...         ethercat.AutomaticInputChan(
        ...             device="servo-drive-key",
        ...             pdo="Position actual value",
        ...             channel=position_ch.key,
        ...         ),
        ...     ],
        ... )
        >>>
        >>> # Configure and start
        >>> client.tasks.configure(read_task)
        >>> read_task.start()

    :param name: A human-readable name for the task.
    :param sample_rate: The rate at which to sample data from slaves (Hz).
    :param stream_rate: The rate at which data is streamed to Synnax (Hz).
    :param data_saving: Whether to save data permanently or just stream it.
    :param auto_start: Whether to start the task automatically when configured.
    :param channels: List of input channel configurations.
    """

    TYPE = "ethercat_read"
    config: ReadTaskConfig
    _internal: task.Task

    def __init__(
        self,
        internal: task.Task | None = None,
        *,
        name: str = "",
        sample_rate: CrudeRate = 0,
        stream_rate: CrudeRate = 0,
        data_saving: bool = True,
        auto_start: bool = False,
        channels: list[InputChan] | None = None,
    ) -> None:
        if internal is not None:
            self._internal = internal
            self.config = ReadTaskConfig.model_validate(internal.config)
            return
        self._internal = task.Task(name=name, type=self.TYPE)
        self.config = ReadTaskConfig(
            sample_rate=sample_rate,
            stream_rate=stream_rate,
            data_saving=data_saving,
            auto_start=auto_start,
            channels=channels if channels is not None else [],
        )


class WriteTask(task.StarterStopperMixin, task.JSONConfigMixin, task.Protocol):
    """A write task for sending commands to EtherCAT slave devices.

    This task configures the Synnax driver to receive commands from Synnax
    channels and write them to EtherCAT slave PDOs during cyclic exchange.

    The task supports state feedback - if state_channel is specified on a
    channel, the current PDO value will be written back to Synnax at the
    configured state_rate.

    Example:
        >>> import synnax as sy
        >>> from synnax import ethercat
        >>>
        >>> # Create command and state channels
        >>> cmd_time = client.channels.create(
        ...     name="cmd_time",
        ...     data_type=sy.DataType.TIMESTAMP,
        ...     is_index=True,
        ... )
        >>> velocity_cmd = client.channels.create(
        ...     name="velocity_cmd",
        ...     data_type=sy.DataType.INT32,
        ...     index=cmd_time.key,
        ... )
        >>> velocity_state = client.channels.create(
        ...     name="velocity_state",
        ...     data_type=sy.DataType.INT32,
        ...     index=cmd_time.key,
        ... )
        >>>
        >>> # Create the write task
        >>> write_task = ethercat.WriteTask(
        ...     name="EtherCAT Velocity Control",
        ...     state_rate=10,
        ...     execution_rate=1000,
        ...     channels=[
        ...         ethercat.AutomaticOutputChan(
        ...             device="servo-drive-key",
        ...             pdo="Target velocity",
        ...             cmd_channel=velocity_cmd.key,
        ...             state_channel=velocity_state.key,
        ...         ),
        ...     ],
        ... )
        >>>
        >>> # Configure and start
        >>> client.tasks.configure(write_task)
        >>> write_task.start()

    :param name: A human-readable name for the task.
    :param state_rate: Rate at which state feedback is written to Synnax (Hz).
    :param execution_rate: Rate at which commands are executed on the bus (Hz).
    :param data_saving: Whether to persist state data to disk.
    :param auto_start: Whether to start the task automatically when configured.
    :param channels: List of output channel configurations.
    """

    TYPE = "ethercat_write"
    config: WriteTaskConfig
    _internal: task.Task

    def __init__(
        self,
        internal: task.Task | None = None,
        *,
        name: str = "",
        state_rate: float = 1.0,
        execution_rate: float = 1000.0,
        data_saving: bool = True,
        auto_start: bool = False,
        channels: list[OutputChan] | None = None,
    ) -> None:
        if internal is not None:
            self._internal = internal
            self.config = WriteTaskConfig.model_validate(internal.config)
            return
        self._internal = task.Task(name=name, type=self.TYPE)
        self.config = WriteTaskConfig(
            state_rate=state_rate,
            execution_rate=execution_rate,
            data_saving=data_saving,
            auto_start=auto_start,
            channels=channels if channels is not None else [],
        )


class Device(device.Device):
    """EtherCAT slave device configuration.

    This class extends the base Device class to provide EtherCAT-specific
    properties including vendor/product identification, bus position, and
    PDO definitions discovered during scanning.

    Typically, EtherCAT devices are created automatically by the driver's
    scan task. However, this class can be used to create devices manually
    or to update device properties.

    Example:
        >>> from synnax import ethercat
        >>>
        >>> # Create an EtherCAT slave device manually
        >>> device = ethercat.Device(
        ...     name="Servo Drive",
        ...     network="eth0",
        ...     position=0,
        ...     vendor_id=0x00000002,
        ...     product_code=0x12345678,
        ...     revision=0x00010000,
        ...     serial=12345,
        ...     rack=rack.key,
        ...     input_pdos=[
        ...         ethercat.PDOEntry(
        ...             name="Position actual value",
        ...             index=0x6064,
        ...             sub_index=0,
        ...             bit_length=32,
        ...             data_type="int32",
        ...         ),
        ...     ],
        ...     output_pdos=[
        ...         ethercat.PDOEntry(
        ...             name="Target velocity",
        ...             index=0x60FF,
        ...             sub_index=0,
        ...             bit_length=32,
        ...             data_type="int32",
        ...         ),
        ...     ],
        ... )
        >>> client.devices.create(device)

    :param name: Human-readable name for the device.
    :param network: Network interface name (e.g., "eth0", "enp3s0").
    :param position: Position of the slave on the EtherCAT bus (0-indexed).
    :param vendor_id: EtherCAT vendor ID from device EEPROM.
    :param product_code: Product code identifying the device model.
    :param revision: Hardware/firmware revision number.
    :param serial: Unique serial number from device EEPROM.
    :param rack: Rack key this device belongs to.
    :param input_pdos: List of input PDO entry definitions (TxPDO, slave->master).
    :param output_pdos: List of output PDO entry definitions (RxPDO, master->slave).
    :param key: Unique key for the device (auto-generated if empty).
    :param configured: Whether the device has been configured.
    :param enabled: Whether the device is enabled for operation.
    """

    def __init__(
        self,
        *,
        name: str = "",
        network: str = "",
        position: int = 0,
        vendor_id: int = 0,
        product_code: int = 0,
        revision: int = 0,
        serial: int = 0,
        rack: int = 0,
        input_pdos: list[PDOEntry] | None = None,
        output_pdos: list[PDOEntry] | None = None,
        key: str = "",
        configured: bool = False,
        enabled: bool = True,
    ):
        """Initialize an EtherCAT slave device."""
        if not key:
            key = str(uuid4())

        props = {
            "network": network,
            "position": position,
            "vendor_id": vendor_id,
            "product_code": product_code,
            "revision": revision,
            "serial": serial,
            "name": name,
            "enabled": enabled,
            "pdos": {
                "inputs": [p.model_dump() for p in (input_pdos or [])],
                "outputs": [p.model_dump() for p in (output_pdos or [])],
            },
        }

        super().__init__(
            key=key,
            location=f"{network}:{position}",
            rack=rack,
            name=name,
            make=MAKE,
            model=MODEL,
            configured=configured,
            properties=props,
        )



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import numpy as np

import synnax as sy

client = sy.Synnax()

data = client.read(sy.TimeRange(1722464238568134144, 1722464246886652416), "press_pt_1")

print("Average", np.average(data))



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import synnax as sy

client = sy.Synnax()

ch = client.channels.create(
    name="strings",
    virtual=True,
    data_type=sy.DataType.STRING,
    retrieve_if_name_exists=True,
)

loop = sy.Loop(sy.Rate.HZ * 20)

i = 0

with client.open_writer(sy.TimeStamp.now(), [ch.key]) as w:
    while loop.wait():
        i += 1
        w.write(ch.key, [f"Sphinx {i}"])



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example demonstrates how to write data to an index channel and its corresponding
data channels in Synnax in a streaming fashion. Streaming data is ideal for live
applications, such as data acquisition from a sensor.
"""

import numpy as np

import synnax as sy

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

# Create an index channel that will be used to store our timestamps.
time_channel = client.channels.create(
    name="stream_write_time",
    is_index=True,
    data_type=sy.DataType.TIMESTAMP,
    retrieve_if_name_exists=True,
)

# Create two data channels that will be used to store our data values. We'll need to
# pass in the key of the time channel to these data channels when they are created
data_channel_1 = client.channels.create(
    name="stream_write_data_1",
    index=time_channel.key,
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
)
data_channel_2 = client.channels.create(
    name="stream_write_data_2",
    index=time_channel.key,
    data_type=sy.DataType.UINT8,
    retrieve_if_name_exists=True,
)

# We'll start our write at the current time. This timestamp should be the same as or
# just before the first timestamp we write.
start = sy.TimeStamp.now()

# The rate at which we'll send samples to the cluster. sy.Loop  is a utility to help
# regulate the timing.
loop = sy.Loop(sy.Rate.HZ * 500)

client.channels.create(
    name="stream_write_avg",
    expression="return (stream_write_data_1 + f32(stream_write_data_2)) / 2",
    retrieve_if_name_exists=True,
)

client.channels.create(
    name="stream_write_avg_squared",
    expression="return stream_write_avg ^ 2",
    retrieve_if_name_exists=True,
)

# Open the writer as a context manager. Using a context manager is recommended as the
# context manager will automatically close the writer when we are done writing. We will
# write to both the time and data channels. To choose the channels to write to, you can
# use either the keys or the names of the channels (here, we're using the keys).
with client.open_writer(
    start,
    [time_channel.key, data_channel_1.key, data_channel_2.key],
) as writer:
    i = 0
    while loop.wait():
        # Write the data to the Synnax cluster using the writer.
        writer.write(
            {
                time_channel.key: sy.TimeStamp.now(),
                data_channel_1.key: np.sin(i / 10) * 25 + 12.5,
                data_channel_2.key: i % 2,
            }
        )
        i += 1



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from freighter.exceptions import Unreachable

from synnax import Synnax
from synnax.cli.flow import Context
from synnax.config import load_options
from synnax.exceptions import AuthError
from synnax.options import Options


def prompt_client_options(ctx: Context) -> Options:
    """Prompts the user for the parameters to connect to a Synnax server.

    :param ctx: The context of the current flow.
    :return: The options to connect to a Synnax server.
    """
    ctx.console.info("Enter your Synnax connection parameters:")
    return Options(
        host=ctx.console.ask("Host", default="localhost"),
        port=ctx.console.ask("Port", default=9090),
        username=ctx.console.ask("Username", default="synnax"),
        password=ctx.console.ask("Password", password=True),
        secure=ctx.console.ask("Secure Connection?", default=False),
    )


def connect_client(ctx: Context) -> Synnax | None:
    """Connects to a Synnax server. Prompts the user for the connection parameters if
    no configuration file exists.

    :param ctx: The context of the current flow.
    :return: The connected Synnax py, or None if the connection failed.
    """
    opts = load_options()
    if opts is None:
        opts = prompt_client_options(ctx)
    else:
        ctx.console.info("Using saved credentials.")
    return connect_from_options(ctx, opts)


def connect_from_options(ctx: Context, opts: Options) -> Synnax | None:
    """Connects to a Synnax server. Prints user-friendly messages to the console if
    the connection fails.

    :param ctx: The context of the current flow.
    :param opts: The options to connect to a Synnax server.
    :return: The connected Synnax py, or None if the connection failed.
    """
    try:
        client = Synnax(**opts.model_dump())
    except Unreachable:
        ctx.console.error(f"Cannot reach Synnax server at {opts.host}:{opts.port}")
        return None
    except AuthError:
        ctx.console.error("Invalid credentials")
        return None
    except Exception as e:
        raise e
    ctx.console.success("Connection successful!")
    return client



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import random

import numpy as np
import pandas as pd
import pytest

import synnax as sy
from synnax.framer.adapter import ReadFrameAdapter, WriteFrameAdapter
from synnax.framer.frame import FramePayload


@pytest.mark.framer
@pytest.mark.telem
class TestFrame:
    def test_construction_from_cols_and_series(self):
        f = sy.Frame(["big", "red", "dog"], [np.array([1, 2, 3]), np.array([4, 5, 6])])
        assert f["big"][0] == 1

    def test_construction_from_dict(self):
        f = sy.Frame({1: sy.Series([1, 2, 3, 4]), 3: sy.Series([4, 5, 6, 7])})
        assert f[1][0] == 1

    def test_construction_from_data_frame(self):
        f = sy.Frame(
            pd.DataFrame(
                {"big": sy.Series([1, 2, 3, 4]), "dog": sy.Series([4, 5, 6, 7])}
            )
        )
        assert f["dog"][0] == 4

    def test_construction_from_keys_and_series(self):
        f = sy.Frame({1: sy.Series([1, 2, 3, 4]), 2: sy.Series([4, 5, 6, 7])})
        assert f[1][0] == 1

    def test_construction_from_frame(self):
        f = sy.Frame({"big": sy.Series([1, 2, 3, 4])})
        f2 = sy.Frame(f)
        assert f is not f2
        assert f.series[0] is f2.series[0]

    def test_construction_from_payload(self):
        pld = FramePayload([1, 2], [sy.Series([1, 2, 3]), sy.Series([4, 5, 6])])
        f = sy.Frame(pld)
        assert f[1][1] == 2

    def test_pandas_interop(self):
        f = sy.Frame({"big": sy.Series([1, 2, 3, 4])})
        assert len(f[f["big"] > 1]["big"]) == 3

    def test_contains_name(self):
        """Should return True if the name is in the Frame"""
        f = sy.Frame({"big": sy.Series([1, 2, 3, 4])})
        assert "big" in f

    def test_contains_key(self):
        """Should return True if the key is in the Frame"""
        f = sy.Frame({122: sy.Series([1, 2, 3, 4])})
        assert 122 in f

    def test_not_contains_name(self):
        """Should return False if the name is not in the Frame"""
        f = sy.Frame({"big": sy.Series([1, 2, 3, 4])})
        assert "small" not in f


@pytest.mark.framer
class TestWriteFrameAdapter:
    @pytest.fixture(scope="class")
    def adapter(self, client: sy.Synnax) -> tuple[WriteFrameAdapter, sy.Channel]:
        ch = client.channels.create(
            name=f"test_{random.randint(0, 100000)}",
            leaseholder=1,
            virtual=True,
            data_type=sy.DataType.FLOAT64,
        )
        adapter = WriteFrameAdapter(client.channels._retriever)
        adapter.update(ch.name)

        return adapter, ch

    def test_adaptation_of_keys_frame(
        self, adapter: tuple[WriteFrameAdapter, sy.Channel]
    ):
        """It should correctly adapt of a Frame keyed by channel key."""
        adapter, ch = adapter
        o = adapter.adapt(
            sy.Frame([ch.key], [sy.Series([1, 2, 3], data_type=sy.DataType.FLOAT64)])
        )
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64

    def test_adaptation_of_names_frame(
        self, adapter: tuple[WriteFrameAdapter, sy.Channel]
    ):
        """It should correctly adapt of a Frame keyed by channel name."""
        adapter, ch = adapter
        o = adapter.adapt(
            sy.Frame([ch.name], [sy.Series([1, 2, 3], data_type=sy.DataType.FLOAT64)])
        )
        assert len(o.channels) == 1
        assert len(o.series) == 1
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64

    def test_adaptation_of_name_series(
        self, adapter: tuple[WriteFrameAdapter, sy.Channel]
    ):
        """It should correctly adapt a first argument of a channel name and a second
        argument of a series."""
        adapter, ch = adapter
        o = adapter.adapt(ch.name, sy.Series([1, 2, 3], data_type=sy.DataType.FLOAT64))
        assert len(o.channels) == 1
        assert len(o.series) == 1
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64

    def test_adaptation_of_name_float(
        self, adapter: tuple[WriteFrameAdapter, sy.Channel]
    ):
        """It should correctly adapt a first argument of a channel name and a second
        argument of a float."""
        adapter, ch = adapter
        o = adapter.adapt(ch.name, 1.0)
        assert len(o.channels) == 1
        assert len(o.series) == 1
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64

    def test_adaptation_of_name_int(
        self, adapter: tuple[WriteFrameAdapter, sy.Channel]
    ):
        """It should correctly adapt a first argument of a channel name and a second
        argument of an int."""
        adapter, ch = adapter
        o = adapter.adapt(ch.name, 1)
        assert len(o.channels) == 1
        assert len(o.series) == 1
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64

    def test_adaptation_of_names_series(
        self, adapter: tuple[WriteFrameAdapter, sy.Channel]
    ):
        """It should correctly adapt a first argument of a channel name and a second
        argument of a series."""
        adapter, ch = adapter
        o = adapter.adapt(
            [ch.name], [sy.Series([1, 2, 3], data_type=sy.DataType.FLOAT64)]
        )
        assert len(o.channels) == 1
        assert len(o.series) == 1
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64

    def test_adaptation_of_dict_series(
        self, adapter: tuple[WriteFrameAdapter, sy.Channel]
    ):
        """It should correctly adapt a dict of channel names to series."""
        adapter, ch = adapter
        o = adapter.adapt({ch.name: sy.Series([1, 2, 3])})
        assert len(o.channels) == 1
        assert len(o.series) == 1
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64

    def test_adaptation_of_dict_float(
        self, adapter: tuple[WriteFrameAdapter, sy.Channel]
    ):
        """It should correctly adapt a dict of channel names to floats."""
        adapter, ch = adapter
        o = adapter.adapt(
            {
                ch.name: 1.0,
            }
        )
        assert len(o.channels) == 1
        assert len(o.series) == 1
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64

    def test_adaptation_of_dict_timestamp(self, client: sy.Synnax):
        """It should correctly adapt a dict of channel names to timestamps."""
        ch = client.channels.create(
            sy.Channel(
                name=f"test_{random.randint(0, 100000)}",
                virtual=True,
                data_type=sy.DataType.TIMESTAMP,
            )
        )
        adapter = WriteFrameAdapter(client.channels._retriever)
        adapter.update(ch.key)
        o = adapter.adapt({ch.name: sy.TimeStamp.now()})
        assert len(o.channels) == 1
        assert len(o.series) == 1
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.TIMESTAMP

    def test_adaptation_of_channel_dict(
        self, adapter: tuple[WriteFrameAdapter, sy.Channel]
    ):
        """It should correctly adapt a dict of channels to series."""
        adapter, ch = adapter
        o = adapter.adapt(
            {
                ch: 1.0,
            }
        )
        assert len(o.channels) == 1
        assert len(o.series) == 1
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64

    def test_adaptation_of_channel_payload(
        self, adapter: tuple[WriteFrameAdapter, sy.Channel]
    ):
        """It should correctly adapt a FramePayload keyed by channel key."""
        adapter, ch = adapter
        o = adapter.adapt(ch, sy.TimeStamp.now())
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64

    def test_adaptation_of_multiple_payloads(
        self,
        adapter: tuple[WriteFrameAdapter, sy.Channel],
    ):
        """Should correctly adapt multiple channels and a single list of values"""
        adapter, ch = adapter
        o = adapter.adapt([ch], [1.0])
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64
        assert o.series[0][0] == 1.0

    def test_adaptation_of_list(
        self,
        adapter: tuple[WriteFrameAdapter, sy.Channel],
    ):
        """Should correctly adapt a channel and a list of values"""
        adapter, ch = adapter
        o = adapter.adapt(ch, [1.0, 2.0, 3.0])
        assert o.channels[0] == ch.key
        assert o.series[0].data_type == sy.DataType.FLOAT64
        assert len(o.series[0]) == 3

    def test_adaptation_of_multiple_series_and_single_payload(self, adapter):
        """Should raise a validation error when there are more series than channels"""
        adapter, ch = adapter
        with pytest.raises(sy.ValidationError):
            adapter.adapt(ch, [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])

    def test_adaptation_of_single_channel_and_no_data(self, adapter):
        """Should raise a validation error when there are no series"""
        adapter, ch = adapter
        with pytest.raises(sy.ValidationError):
            adapter.adapt(ch)

    def test_adaptation_of_multiple_channels_and_no_data(self, adapter):
        """Should raise a validation error when there are no series"""
        adapter, ch = adapter
        with pytest.raises(sy.ValidationError):
            adapter.adapt([ch, ch])

    def test_mismatch_of_channels_and_series_length(self, adapter):
        """Should raise a validation error when there are more channels than series"""
        adapter, ch = adapter
        with pytest.raises(sy.ValidationError):
            adapter.adapt([ch, ch], [1.0])

    def test_validation_error_when_frame_with_nonexistent_channel_name_is_adapted(
        self, adapter
    ):
        """Should raise a validation error when a Frame with a nonexistent channel
        key is adapted
        """
        adapter, _ = adapter
        with pytest.raises(sy.ValidationError):
            adapter.adapt(
                sy.Frame(
                    ["caramela"], [sy.Series([1, 2, 3], data_type=sy.DataType.FLOAT64)]
                )
            )

    def test_type_error_when_invalid_value_is_adapted(self, adapter):
        """Should raise a type error when an invalid value is adapted"""
        adapter, channel = adapter
        with pytest.raises(TypeError):
            adapter.adapt(Exception("Invalid value"))

    def test_adapt_dict_keys(self, adapter):
        """Should raise a type error when an invalid value is adapted"""
        adapter, ch = adapter
        adapted = adapter.adapt_dict_keys({ch.name: 123})
        assert adapted[ch.key] == 123

    def test_adapt_single_string(self, client):
        """Should correctly adapt a single string into a string based series"""
        ch = client.channels.create(
            name=f"test_{random.randint(0, 100000)}",
            virtual=True,
            data_type=sy.DataType.STRING,
        )
        adapter = WriteFrameAdapter(client.channels._retriever)
        adapter.update(ch.name)
        adapted = adapter.adapt({ch.name: "hello"})
        assert adapted[ch.key][0] == "hello"

    def test_adapt_single_string_name_value_pair(self, client):
        ch = client.channels.create(
            name=f"test_{random.randint(0, 100000)}",
            virtual=True,
            data_type=sy.DataType.STRING,
        )
        adapter = WriteFrameAdapter(client.channels._retriever)
        adapter.update(ch.name)
        adapted = adapter.adapt(ch.name, "hello")
        assert adapted[ch.key][0] == "hello"


@pytest.mark.framer
class TestReadFrameAdapter:
    """Comprehensive tests for ReadFrameAdapter functionality."""

    @pytest.fixture(scope="class")
    def channels(self, client: sy.Synnax) -> tuple[sy.Channel, sy.Channel, sy.Channel]:
        """Create three test channels for adapter testing."""
        ch1 = client.channels.create(
            name=f"test_read_adapter_ch1_{random.randint(0, 100000)}",
            virtual=True,
            data_type=sy.DataType.FLOAT64,
        )
        ch2 = client.channels.create(
            name=f"test_read_adapter_ch2_{random.randint(0, 100000)}",
            virtual=True,
            data_type=sy.DataType.INT64,
        )
        ch3 = client.channels.create(
            name=f"test_read_adapter_ch3_{random.randint(0, 100000)}",
            virtual=True,
            data_type=sy.DataType.FLOAT32,
        )
        return ch1, ch2, ch3

    def test_initialization(self, client: sy.Synnax):
        """Should successfully initialize a ReadFrameAdapter."""
        adapter = ReadFrameAdapter(client.channels._retriever)
        assert adapter.retriever is not None
        assert adapter.keys == []
        assert adapter.codec is not None

    def test_update_with_keys(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should update adapter with channel keys without creating name mapping."""
        ch1, ch2, _ = channels
        adapter = ReadFrameAdapter(client.channels._retriever)
        adapter.update([ch1.key, ch2.key])

        # Should store keys directly
        assert adapter.keys == [ch1.key, ch2.key]

    def test_update_with_names(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should update adapter with channel names and create key->name mapping."""
        ch1, ch2, _ = channels
        adapter = ReadFrameAdapter(client.channels._retriever)
        adapter.update([ch1.name, ch2.name])

        # Should store keys (not names)
        assert adapter.keys == [ch1.key, ch2.key]

    def test_update_with_nonexistent_channel(self, client: sy.Synnax):
        """Should raise KeyError when updating with a nonexistent channel name."""
        adapter = ReadFrameAdapter(client.channels._retriever)
        with pytest.raises(KeyError):
            adapter.update(["nonexistent_channel_name_12345"])

    def test_adapt_with_keys_no_conversion(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should return frame unchanged when adapter was updated with keys."""
        ch1, ch2, _ = channels
        adapter = ReadFrameAdapter(client.channels._retriever)
        adapter.update([ch1.key, ch2.key])

        # Create a frame with keys
        original_frame = sy.Frame(
            [ch1.key, ch2.key],
            [
                sy.Series([1.0, 2.0, 3.0], data_type=sy.DataType.FLOAT64),
                sy.Series([10, 20, 30], data_type=sy.DataType.INT64),
            ],
        )

        adapted_frame = adapter.adapt(original_frame)

        # Frame should be returned unchanged
        assert adapted_frame.channels == [ch1.key, ch2.key]
        assert len(adapted_frame.series) == 2
        assert np.array_equal(adapted_frame[ch1.key], np.array([1.0, 2.0, 3.0]))
        assert np.array_equal(adapted_frame[ch2.key], np.array([10, 20, 30]))

    def test_adapt_with_names_conversion(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should convert channel keys to names when adapter was updated with names."""
        ch1, ch2, _ = channels
        adapter = ReadFrameAdapter(client.channels._retriever)
        adapter.update([ch1.name, ch2.name])

        # Create a frame with keys (as would come from server)
        frame_with_keys = sy.Frame(
            [ch1.key, ch2.key],
            [
                sy.Series([1.0, 2.0, 3.0], data_type=sy.DataType.FLOAT64),
                sy.Series([10, 20, 30], data_type=sy.DataType.INT64),
            ],
        )

        adapted_frame = adapter.adapt(frame_with_keys)

        # Keys should be converted to names
        assert adapted_frame.channels == [ch1.name, ch2.name]
        assert len(adapted_frame.series) == 2
        assert np.array_equal(adapted_frame[ch1.name], np.array([1.0, 2.0, 3.0]))
        assert np.array_equal(adapted_frame[ch2.name], np.array([10, 20, 30]))

    def test_adapt_filters_extra_channels(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should filter out channels not in the adapter mapping."""
        ch1, ch2, ch3 = channels
        adapter = ReadFrameAdapter(client.channels._retriever)
        # Only register ch1 and ch2
        adapter.update([ch1.name, ch2.name])

        # Create frame with all three channels
        frame_with_extra = sy.Frame(
            [ch1.key, ch2.key, ch3.key],
            [
                sy.Series([1.0, 2.0], data_type=sy.DataType.FLOAT64),
                sy.Series([10, 20], data_type=sy.DataType.INT64),
                sy.Series([100.0, 200.0], data_type=sy.DataType.FLOAT32),
            ],
        )

        adapted_frame = adapter.adapt(frame_with_extra)

        # Only ch1 and ch2 should remain
        assert len(adapted_frame.channels) == 2
        assert len(adapted_frame.series) == 2
        assert ch1.name in adapted_frame.channels
        assert ch2.name in adapted_frame.channels
        assert ch3.name not in adapted_frame.channels
        # Data should be preserved for kept channels
        assert np.array_equal(adapted_frame[ch1.name], np.array([1.0, 2.0]))
        assert np.array_equal(adapted_frame[ch2.name], np.array([10, 20]))

    def test_adapt_with_empty_frame(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should handle empty frames gracefully."""
        ch1, ch2, _ = channels
        adapter = ReadFrameAdapter(client.channels._retriever)
        adapter.update([ch1.name, ch2.name])

        # Create empty frame
        empty_frame = sy.Frame([], [])

        adapted_frame = adapter.adapt(empty_frame)

        assert len(adapted_frame.channels) == 0
        assert len(adapted_frame.series) == 0

    def test_adapt_with_no_matching_channels(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should return empty frame when no channels match the adapter mapping."""
        ch1, ch2, ch3 = channels
        adapter = ReadFrameAdapter(client.channels._retriever)
        # Register ch1 and ch2
        adapter.update([ch1.name, ch2.name])

        # Create frame with only ch3 (not in adapter)
        frame_no_match = sy.Frame(
            [ch3.key],
            [sy.Series([1.0, 2.0, 3.0], data_type=sy.DataType.FLOAT32)],
        )

        adapted_frame = adapter.adapt(frame_no_match)

        # All channels should be filtered out
        assert len(adapted_frame.channels) == 0
        assert len(adapted_frame.series) == 0

    def test_adapt_with_partial_matches(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should keep only matching channels and filter non-matching ones."""
        ch1, ch2, ch3 = channels
        adapter = ReadFrameAdapter(client.channels._retriever)
        adapter.update([ch1.name, ch3.name])  # Register ch1 and ch3

        # Frame contains ch1, ch2, ch3 but adapter only has ch1 and ch3
        frame_partial = sy.Frame(
            [ch1.key, ch2.key, ch3.key],
            [
                sy.Series([1.0, 2.0], data_type=sy.DataType.FLOAT64),
                sy.Series([10, 20], data_type=sy.DataType.INT64),
                sy.Series([100.0, 200.0], data_type=sy.DataType.FLOAT32),
            ],
        )

        adapted_frame = adapter.adapt(frame_partial)

        # Should only have ch1 and ch3
        assert len(adapted_frame.channels) == 2
        assert ch1.name in adapted_frame.channels
        assert ch3.name in adapted_frame.channels
        assert ch2.name not in adapted_frame.channels
        # Verify data integrity
        assert np.array_equal(adapted_frame[ch1.name], np.array([1.0, 2.0]))
        assert np.array_equal(adapted_frame[ch3.name], np.array([100.0, 200.0]))

    def test_adapt_preserves_series_data_types(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should preserve series data types during adaptation."""
        ch1, ch2, _ = channels
        adapter = ReadFrameAdapter(client.channels._retriever)
        adapter.update([ch1.name, ch2.name])

        frame = sy.Frame(
            [ch1.key, ch2.key],
            [
                sy.Series([1.0, 2.0, 3.0], data_type=sy.DataType.FLOAT64),
                sy.Series([10, 20, 30], data_type=sy.DataType.INT64),
            ],
        )

        adapted_frame = adapter.adapt(frame)

        # Data types should be preserved
        assert adapted_frame[ch1.name].data_type == sy.DataType.FLOAT64
        assert adapted_frame[ch2.name].data_type == sy.DataType.INT64

    def test_adapt_with_mixed_keys_and_names(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should handle frames with mixed keys and names (though unusual)."""
        ch1, ch2, _ = channels
        adapter = ReadFrameAdapter(client.channels._retriever)
        adapter.update([ch1.name, ch2.name])

        # Frame with mix of key and name (ch1 as key, ch2 as name)
        # This tests the isinstance(k, ChannelKey) check in adapt method
        frame_mixed = sy.Frame(
            [ch1.key, ch2.name],
            [
                sy.Series([1.0, 2.0], data_type=sy.DataType.FLOAT64),
                sy.Series([10, 20], data_type=sy.DataType.INT64),
            ],
        )

        adapted_frame = adapter.adapt(frame_mixed)

        # Both should be converted to names
        assert ch1.name in adapted_frame.channels
        # ch2.name should remain unchanged since it's already a name
        assert ch2.name in adapted_frame.channels

    def test_adapt_preserves_series_order(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should preserve the order of series when adapting."""
        ch1, ch2, ch3 = channels
        adapter = ReadFrameAdapter(client.channels._retriever)
        adapter.update([ch1.name, ch2.name, ch3.name])

        frame = sy.Frame(
            [ch1.key, ch2.key, ch3.key],
            [
                sy.Series([1.0, 2.0], data_type=sy.DataType.FLOAT64),
                sy.Series([10, 20], data_type=sy.DataType.INT64),
                sy.Series([100.0, 200.0], data_type=sy.DataType.FLOAT32),
            ],
        )

        adapted_frame = adapter.adapt(frame)

        # Order should be preserved
        assert adapted_frame.channels[0] == ch1.name
        assert adapted_frame.channels[1] == ch2.name
        assert adapted_frame.channels[2] == ch3.name

    def test_codec_update(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should update codec with channel keys and data types."""
        ch1, ch2, _ = channels
        adapter = ReadFrameAdapter(client.channels._retriever)

        # Codec should be empty initially
        initial_codec = adapter.codec

        # Update with channels
        adapter.update([ch1.name, ch2.name])

        # Codec should have been updated
        assert adapter.codec is initial_codec  # Same instance
        # Keys should be registered in adapter
        assert ch1.key in adapter.keys
        assert ch2.key in adapter.keys

    def test_multiple_updates(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should handle multiple update calls correctly."""
        ch1, ch2, ch3 = channels
        adapter = ReadFrameAdapter(client.channels._retriever)

        # First update with ch1 and ch2
        adapter.update([ch1.name, ch2.name])
        assert len(adapter.keys) == 2

        # Second update with ch3 only
        adapter.update([ch3.name])
        assert len(adapter.keys) == 1
        assert adapter.keys == [ch3.key]

        # Adapt should now only work for ch3
        frame = sy.Frame(
            [ch1.key, ch2.key, ch3.key],
            [
                sy.Series([1.0], data_type=sy.DataType.FLOAT64),
                sy.Series([10], data_type=sy.DataType.INT64),
                sy.Series([100.0], data_type=sy.DataType.FLOAT32),
            ],
        )

        adapted_frame = adapter.adapt(frame)
        assert len(adapted_frame.channels) == 1
        assert adapted_frame.channels[0] == ch3.name

    def test_update_switches_between_keys_and_names(
        self, client: sy.Synnax, channels: tuple[sy.Channel, sy.Channel, sy.Channel]
    ):
        """Should correctly switch between key mode and name mode."""
        ch1, ch2, _ = channels
        adapter = ReadFrameAdapter(client.channels._retriever)

        # Start with names
        adapter.update([ch1.name, ch2.name])
        frame_names = sy.Frame(
            [ch1.key, ch2.key],
            [
                sy.Series([1.0, 2.0], data_type=sy.DataType.FLOAT64),
                sy.Series([10, 20], data_type=sy.DataType.INT64),
            ],
        )
        adapted_names = adapter.adapt(frame_names)
        assert adapted_names.channels == [ch1.name, ch2.name]

        # Switch to keys
        adapter.update([ch1.key, ch2.key])
        frame_keys = sy.Frame(
            [ch1.key, ch2.key],
            [
                sy.Series([3.0, 4.0], data_type=sy.DataType.FLOAT64),
                sy.Series([30, 40], data_type=sy.DataType.INT64),
            ],
        )
        adapted_keys = adapter.adapt(frame_keys)
        # Should return unchanged (keys)
        assert adapted_keys.channels == [ch1.key, ch2.key]



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example demonstrates how to read thermocouple data from a LabJack device.

Before running this example:
1. Connect a LabJack T7 or T7-Pro to your computer

2. Register the device in Synnax:
   uv run python examples/labjack/connect_device.py

3. Connect thermocouples to AIN0 and AIN2 (or update channel configuration below)

4. Run this script:
   uv run python examples/labjack/thermocouple_read_task.py

This example reads from two K-type thermocouples with device CJC at 10 Hz.
Note: Thermocouple streaming is slower than standard analog inputs due to the
      required cold junction compensation calculations.
"""

import synnax as sy

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

# Retrieve the LabJack device from Synnax
# Update this with the name you gave the device in the Synnax Console
dev = client.devices.retrieve(name="My LabJack T7")

# Create an index channel that will be used to store the timestamps for the data.
labjack_tc_time = client.channels.create(
    name="labjack_tc_time",
    is_index=True,
    data_type=sy.DataType.TIMESTAMP,
    retrieve_if_name_exists=True,
)

# Create Synnax channels to store the thermocouple temperature data.
tc0 = client.channels.create(
    name="thermocouple_0",
    index=labjack_tc_time.key,
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
)

tc1 = client.channels.create(
    name="thermocouple_1",
    index=labjack_tc_time.key,
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
)

# Create the LabJack Read Task with Thermocouples
# Reads two K-type thermocouples at 10 Hz with device cold junction compensation
tsk = sy.labjack.ReadTask(
    name="LabJack Py - Thermocouple Task",
    device=dev.key,
    sample_rate=sy.Rate.HZ * 10,  # Sample at 10 Hz (max for thermocouples)
    stream_rate=sy.Rate.HZ * 10,  # Stream at 10 Hz
    data_saving=True,
    channels=[
        # K-type thermocouple on AIN0 with device CJC in Celsius
        sy.labjack.ThermocoupleChan(
            port="AIN0",
            channel=tc0.key,
            thermocouple_type="K",
            cjc_source="TEMPERATURE_DEVICE_K",  # Use device internal temp sensor
            cjc_slope=1.0,  # Device temp sensor
            cjc_offset=0.0,  # Device temp sensor
            units="C",  # Celsius
            pos_chan=0,
            neg_chan=199,  # 199 = single-ended (GND)
        ),
        # K-type thermocouple on AIN2 with device CJC in Fahrenheit
        sy.labjack.ThermocoupleChan(
            port="AIN2",
            channel=tc1.key,
            thermocouple_type="K",
            cjc_source="TEMPERATURE_DEVICE_K",  # Use device internal temp sensor
            cjc_slope=1.0,  # Device temp sensor
            cjc_offset=0.0,  # Device temp sensor
            units="F",  # Fahrenheit
            pos_chan=2,
            neg_chan=199,  # 199 = single-ended (GND)
        ),
    ],
)

# Configure the task with Synnax
try:
    client.tasks.configure(tsk)
    print("✓ Task configured successfully")
except Exception as e:
    print(f"✗ Task configuration failed: {e}")
    exit(1)

print("=" * 70)
print("Starting LabJack Thermocouple Read Task")
print("=" * 70)
print("Reading K-type thermocouples from AIN0 and AIN2...")
print("Running continuously - Press Ctrl+C to stop\n")

print(f"{'Sample':<10} {'Time (s)':<10} {'TC0 (°C)':>12} {'TC1 (°F)':>12}")
print("-" * 70)

# Start the task and read data continuously
try:
    # Hide cursor for clean output
    print("\033[?25l", end="", flush=True)

    with tsk.run():
        with client.open_streamer(["thermocouple_0", "thermocouple_1"]) as streamer:
            sample_count = 0
            start_time = sy.TimeStamp.now()

            while True:
                frame = streamer.read()
                if frame:
                    # Print the latest values from both channels
                    if "thermocouple_0" in frame and len(frame["thermocouple_0"]) > 0:
                        temp_c = frame["thermocouple_0"][-1]
                        temp_f = frame["thermocouple_1"][-1]
                        elapsed = sy.TimeStamp.now().span(start_time).seconds

                        sample_count += 1
                        print(
                            f"{sample_count:<10} {elapsed:<10.1f} {temp_c:>12.1f} {temp_f:>12.1f}",
                            end="\r",
                            flush=True,
                        )

# Output summary
except KeyboardInterrupt:
    print("\n" + "-" * 70)
    print("✓ Thermocouple read task stopped by user")
    print(f"\nCollected {sample_count} samples")
    print("\nNote: Thermocouple data is acquired using unary reads (not streaming)")
    print("      due to LJM limitations with extended features.")
    print("=" * 70)
finally:
    # Ensure cursor is always shown even if something goes wrong
    print("\033[?25h", end="", flush=True)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from uuid import uuid4 as uuid

import pytest

import synnax as sy


@pytest.fixture
def new_user():
    return sy.user.New(
        username=uuid().__str__(),
        password="password",
        first_name="first",
        last_name="last",
        key=uuid(),
    )


@pytest.fixture
def new_user_list():
    return [
        sy.user.New(
            username=uuid().__str__(),
            password="password1",
            first_name="first1",
            last_name="last1",
        ),
        sy.user.New(
            username=uuid().__str__(),
            password="password2",
            first_name="first2",
            key=uuid(),
        ),
        sy.user.New(username=uuid().__str__(), password="password3", key=uuid()),
        sy.user.New(username=uuid().__str__(), password="password4"),
    ]


def compare_users(user: sy.user.User, new_user: sy.user.New):
    assert user.username == new_user.username
    assert user.first_name == new_user.first_name
    assert user.last_name == new_user.last_name
    if new_user.key is not None:
        assert user.key == new_user.key
    else:
        assert user.key is not None


class TestUserClient:
    def test_create_params(self, client: sy.Synnax):
        user = client.users.create(username="test", password="test")
        compare_users(user, sy.user.New(username="test", password="test"))
        client.users.delete(user.key)
        key = uuid()
        username = uuid().__str__()
        user = client.users.create(
            username=username, password="test", first_name="silly", key=key
        )
        compare_users(
            user,
            sy.user.New(
                username=username, password="test", first_name="silly", key=key
            ),
        )
        client.users.delete(user.key)

    def test_create_user(self, client: sy.Synnax, new_user):
        repeated_user = client.users.create(user=new_user)
        compare_users(repeated_user, new_user)

    def test_create_repeated(self, client: sy.Synnax):
        with pytest.raises(sy.AuthError):
            client.users.create(username="synnax", password="test")

    def test_create_many(self, client: sy.Synnax, new_user_list):
        repeated_user_list = client.users.create(users=new_user_list)
        assert len(repeated_user_list) == len(new_user_list)
        for i in range(len(repeated_user_list)):
            compare_users(repeated_user_list[i], new_user_list[i])



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""OPC UA example package.

Lazy imports avoid the RuntimeWarning when running
``python -m examples.opcua.server``.
"""

__all__ = ["run_server", "OPCUASim", "OPCUATLSSim", "OPCUATLSAuthSim"]


def __getattr__(name: str):
    if name in __all__:
        from . import server

        return getattr(server, name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
Diagnostic script to test Modbus connection and task status.
"""

import synnax as sy

# Connect to Synnax
client = sy.Synnax()

print("=" * 70)
print("Modbus Connection Diagnostic")
print("=" * 70)

# Check if device exists
dev = client.devices.retrieve(name="Modbus Server", ignore_not_found=True)

if dev is None:
    print("✗ Device 'Modbus Server' not found")
    print("  Run connect_modbus_server.py first")
    exit(1)

print(f"✓ Device found: {dev.name}")
print(f"  - Key: {dev.key}")
print(f"  - Location: {dev.location}")
print(f"  - Make: {dev.make}")
print(f"  - Properties: {dev.properties}")
print()

# Check if any tasks exist for this device
print("Checking for tasks...")
try:
    # Retrieve task by name
    task = client.tasks.retrieve(name="Modbus Python Example - Read Task")
    print(f"  ✓ Found task: {task.name}")
    print(f"    - Key: {task.key}")
    print(f"    - Type: {task.type}")
    if task.status:
        print(f"    - Status: {task.status.variant}")
        if task.status.message:
            print(f"    - Message: {task.status.message}")
except Exception as e:
    print(f"  ✗ Task not found or error: {e}")

print()

# Check if channels exist
print("Checking for channels...")
expected_channels = [
    "modbus_time",
    "holding_register_0",
    "holding_register_1",
    "input_register_0",
    "coil_0",
    "discrete_input_0",
]

for ch_name in expected_channels:
    try:
        ch = client.channels.retrieve(ch_name)
        print(f"  ✓ {ch_name} (key={ch.key})")
    except Exception:
        print(f"  ✗ {ch_name} - not found")

print()
print("=" * 70)
print("Diagnostic complete")
print()
print("Next steps:")
print("1. Make sure the Modbus server is running:")
print("   uv run python driver/modbus/dev/server.py")
print("2. Check the Synnax Console for any error messages")
print("3. Check the driver logs for connection errors")
print("=" * 70)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import time

import numpy as np

import synnax as sy

client = sy.Synnax()

idx = client.channels.create(
    name="idx", is_index=True, data_type="timestamp", retrieve_if_name_exists=True
)

data = client.channels.create(
    name="data", index=idx.key, data_type="float32", retrieve_if_name_exists=True
)

count = 100
with client.open_writer(sy.TimeStamp.now(), [idx.key, data.key]) as writer:
    for i in range(count):
        time.sleep(0.1)
        print(np.round(i / count * count))
        writer.write(
            {
                idx.key: sy.TimeStamp.now(),
                data.key: np.sin(i),
            }
        )


time.sleep(1)

data_2 = client.channels.create(
    name="data_2", index=idx.key, data_type="float32", retrieve_if_name_exists=True
)

data_3 = client.channels.create(
    name="data_3", index=idx.key, data_type="float32", retrieve_if_name_exists=True
)

with client.open_writer(
    sy.TimeStamp.now(),
    [data.key, data_2.key, data_3.key, idx.key],
) as writer:
    for i in range(50000):
        time.sleep(0.1)
        writer.write(
            {
                idx.key: sy.TimeStamp.now(),
                data.key: np.sin(i),
                data_2.key: np.sin(i * 2),
                data_3.key: np.sin(i * 3),
            }
        )
